#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This package is deprecated. Please use :mod:`airflow.utils`."""
from __future__ import annotations

import warnings

from airflow.exceptions import RemovedInAirflow3Warning
from airflow.utils.deprecation_tools import add_deprecated_classes

warnings.warn(
    "This module is deprecated. Please use `airflow.utils`.",
    RemovedInAirflow3Warning,
    stacklevel=2
)

__deprecated_classes = {
    "gcp_field_sanitizer": {
        "GcpBodyFieldSanitizer": "airflow.providers.google.cloud.utils.field_sanitizer.GcpBodyFieldSanitizer",
        "GcpFieldSanitizerException": (
            "airflow.providers.google.cloud.utils.field_sanitizer.GcpFieldSanitizerException"
        ),
    },
    "gcp_field_validator": {
        "GcpBodyFieldValidator": "airflow.providers.google.cloud.utils.field_validator.GcpBodyFieldValidator",
        "GcpFieldValidationException": (
            "airflow.providers.google.cloud.utils.field_validator.GcpFieldValidationException"
        ),
        "GcpValidationSpecificationException": (
            "airflow.providers.google.cloud.utils.field_validator.GcpValidationSpecificationException"
        ),
    },
    "mlengine_operator_utils": {
        "create_evaluate_ops": (
            "airflow.providers.google.cloud.utils.mlengine_operator_utils.create_evaluate_ops"
        ),
    },
    "mlengine_prediction_summary": {
        "JsonCoder": "airflow.providers.google.cloud.utils.mlengine_prediction_summary.JsonCoder",
        "MakeSummary": "airflow.providers.google.cloud.utils.mlengine_prediction_summary.MakeSummary",
    },
    "sendgrid": {
        "import_string": "airflow.utils.module_loading.import_string",
    },
    "weekday": {
        "WeekDay": "airflow.utils.weekday.WeekDay",
    },
}

add_deprecated_classes(__deprecated_classes, __name__)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import NamedTuple

from marshmallow import Schema, fields
from marshmallow_sqlalchemy import SQLAlchemySchema, auto_field

from airflow.models.dagwarning import DagWarning


class DagWarningSchema(SQLAlchemySchema):
    """Import error schema."""

    class Meta:
        """Meta."""

        model = DagWarning

    dag_id = auto_field(data_key="dag_id", dump_only=True)
    warning_type = auto_field()
    message = auto_field()
    timestamp = auto_field(format="iso")


class DagWarningCollection(NamedTuple):
    """List of dag warnings with metadata."""

    dag_warnings: list[DagWarning]
    total_entries: int


class DagWarningCollectionSchema(Schema):
    """Import error collection schema."""

    dag_warnings = fields.List(fields.Nested(DagWarningSchema))
    total_entries = fields.Int()


dag_warning_schema = DagWarningSchema()
dag_warning_collection_schema = DagWarningCollectionSchema()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from deprecated import deprecated

from airflow.auth.managers.fab.security_manager.override import FabAirflowSecurityManagerOverride

EXISTING_ROLES = {
    "Admin",
    "Viewer",
    "User",
    "Op",
    "Public",
}


@deprecated(
    reason="If you want to override the security manager, you should inherit from "
    "`airflow.auth.managers.fab.security_manager.override.FabAirflowSecurityManagerOverride` "
    "instead"
)
class AirflowSecurityManager(FabAirflowSecurityManagerOverride):
    """Placeholder, just here to avoid breaking the code of users who inherit from this.

    Do not use if writing new code.
    """

    ...



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This package is deprecated. Please use `airflow.operators` or `airflow.providers.*.operators`."""
from __future__ import annotations

import warnings

from airflow.exceptions import RemovedInAirflow3Warning
from airflow.utils.deprecation_tools import add_deprecated_classes

warnings.warn(
    "This package is deprecated. Please use `airflow.operators` or `airflow.providers.*.operators`.",
    RemovedInAirflow3Warning,
    stacklevel=2,
)

__deprecated_classes = {
    "adls_list_operator": {
        "ADLSListOperator": "airflow.providers.microsoft.azure.operators.adls.ADLSListOperator",
        "AzureDataLakeStorageListOperator": (
            "airflow.providers.microsoft.azure.operators.adls.ADLSListOperator"
        ),
    },
    "adls_to_gcs": {
        "ADLSToGCSOperator": "airflow.providers.google.cloud.transfers.adls_to_gcs.ADLSToGCSOperator",
        "AdlsToGoogleCloudStorageOperator": (
            "airflow.providers.google.cloud.transfers.adls_to_gcs.ADLSToGCSOperator"
        ),
    },
    "aws_athena_operator": {
        "AWSAthenaOperator": "airflow.providers.amazon.aws.operators.athena.AthenaOperator",
    },
    "aws_sqs_publish_operator": {
        "SqsPublishOperator": "airflow.providers.amazon.aws.operators.sqs.SqsPublishOperator",
        "SQSPublishOperator": "airflow.providers.amazon.aws.operators.sqs.SqsPublishOperator",
    },
    "awsbatch_operator": {
        "BatchProtocol": "airflow.providers.amazon.aws.hooks.batch_client.BatchProtocol",
        "BatchOperator": "airflow.providers.amazon.aws.operators.batch.BatchOperator",
        "AWSBatchOperator": "airflow.providers.amazon.aws.operators.batch.BatchOperator",
    },
    "azure_container_instances_operator": {
        "AzureContainerInstancesOperator": (
            "airflow.providers.microsoft.azure.operators.container_instances.AzureContainerInstancesOperator"
        ),
    },
    "azure_cosmos_operator": {
        "AzureCosmosInsertDocumentOperator": (
            "airflow.providers.microsoft.azure.operators.cosmos.AzureCosmosInsertDocumentOperator"
        ),
    },
    "bigquery_check_operator": {
        "BigQueryCheckOperator": "airflow.providers.google.cloud.operators.bigquery.BigQueryCheckOperator",
        "BigQueryIntervalCheckOperator": (
            "airflow.providers.google.cloud.operators.bigquery.BigQueryIntervalCheckOperator"
        ),
        "BigQueryValueCheckOperator": (
            "airflow.providers.google.cloud.operators.bigquery.BigQueryValueCheckOperator"
        ),
    },
    "bigquery_get_data": {
        "BigQueryGetDataOperator": (
            "airflow.providers.google.cloud.operators.bigquery.BigQueryGetDataOperator"
        ),
    },
    "bigquery_operator": {
        "BigQueryCreateEmptyDatasetOperator": (
            "airflow.providers.google.cloud.operators.bigquery.BigQueryCreateEmptyDatasetOperator"
        ),
        "BigQueryCreateEmptyTableOperator": (
            "airflow.providers.google.cloud.operators.bigquery.BigQueryCreateEmptyTableOperator"
        ),
        "BigQueryCreateExternalTableOperator": (
            "airflow.providers.google.cloud.operators.bigquery.BigQueryCreateExternalTableOperator"
        ),
        "BigQueryDeleteDatasetOperator": (
            "airflow.providers.google.cloud.operators.bigquery.BigQueryDeleteDatasetOperator"
        ),
        "BigQueryExecuteQueryOperator": (
            "airflow.providers.google.cloud.operators.bigquery.BigQueryExecuteQueryOperator"
        ),
        "BigQueryGetDatasetOperator": (
            "airflow.providers.google.cloud.operators.bigquery.BigQueryGetDatasetOperator"
        ),
        "BigQueryGetDatasetTablesOperator": (
            "airflow.providers.google.cloud.operators.bigquery.BigQueryGetDatasetTablesOperator"
        ),
        "BigQueryPatchDatasetOperator": (
            "airflow.providers.google.cloud.operators.bigquery.BigQueryPatchDatasetOperator"
        ),
        "BigQueryUpdateDatasetOperator": (
            "airflow.providers.google.cloud.operators.bigquery.BigQueryUpdateDatasetOperator"
        ),
        "BigQueryUpsertTableOperator": (
            "airflow.providers.google.cloud.operators.bigquery.BigQueryUpsertTableOperator"
        ),
        "BigQueryOperator": "airflow.providers.google.cloud.operators.bigquery.BigQueryExecuteQueryOperator",
    },
    "bigquery_table_delete_operator": {
        "BigQueryDeleteTableOperator": (
            "airflow.providers.google.cloud.operators.bigquery.BigQueryDeleteTableOperator"
        ),
        "BigQueryTableDeleteOperator": (
            "airflow.providers.google.cloud.operators.bigquery.BigQueryDeleteTableOperator"
        ),
    },
    "bigquery_to_bigquery": {
        "BigQueryToBigQueryOperator": (
            "airflow.providers.google.cloud.transfers.bigquery_to_bigquery.BigQueryToBigQueryOperator"
        ),
    },
    "bigquery_to_gcs": {
        "BigQueryToGCSOperator": (
            "airflow.providers.google.cloud.transfers.bigquery_to_gcs.BigQueryToGCSOperator"
        ),
        "BigQueryToCloudStorageOperator": (
            "airflow.providers.google.cloud.transfers.bigquery_to_gcs.BigQueryToGCSOperator"
        ),
    },
    "bigquery_to_mysql_operator": {
        "BigQueryToMySqlOperator": (
            "airflow.providers.google.cloud.transfers.bigquery_to_mysql.BigQueryToMySqlOperator"
        ),
    },
    "cassandra_to_gcs": {
        "CassandraToGCSOperator": (
            "airflow.providers.google.cloud.transfers.cassandra_to_gcs.CassandraToGCSOperator"
        ),
        "CassandraToGoogleCloudStorageOperator": (
            "airflow.providers.google.cloud.transfers.cassandra_to_gcs.CassandraToGCSOperator"
        ),
    },
    "databricks_operator": {
        "DatabricksRunNowOperator": (
            "airflow.providers.databricks.operators.databricks.DatabricksRunNowOperator"
        ),
        "DatabricksSubmitRunOperator": (
            "airflow.providers.databricks.operators.databricks.DatabricksSubmitRunOperator"
        ),
    },
    "dataflow_operator": {
        "DataflowCreateJavaJobOperator": (
            "airflow.providers.google.cloud.operators.dataflow.DataflowCreateJavaJobOperator"
        ),
        "DataflowCreatePythonJobOperator": (
            "airflow.providers.google.cloud.operators.dataflow.DataflowCreatePythonJobOperator"
        ),
        "DataflowTemplatedJobStartOperator": (
            "airflow.providers.google.cloud.operators.dataflow.DataflowTemplatedJobStartOperator"
        ),
        "DataFlowJavaOperator": (
            "airflow.providers.google.cloud.operators.dataflow.DataflowCreateJavaJobOperator"
        ),
        "DataFlowPythonOperator": (
            "airflow.providers.google.cloud.operators.dataflow.DataflowCreatePythonJobOperator"
        ),
        "DataflowTemplateOperator": (
            "airflow.providers.google.cloud.operators.dataflow.DataflowTemplatedJobStartOperator"
        ),
    },
    "dataproc_operator": {
        "DataprocCreateClusterOperator": (
            "airflow.providers.google.cloud.operators.dataproc.DataprocCreateClusterOperator"
        ),
        "DataprocDeleteClusterOperator": (
            "airflow.providers.google.cloud.operators.dataproc.DataprocDeleteClusterOperator"
        ),
        "DataprocInstantiateInlineWorkflowTemplateOperator":
            "airflow.providers.google.cloud.operators.dataproc."
            "DataprocInstantiateInlineWorkflowTemplateOperator",
        "DataprocInstantiateWorkflowTemplateOperator": (
            "airflow.providers.google.cloud.operators.dataproc.DataprocInstantiateWorkflowTemplateOperator"
        ),
        "DataprocJobBaseOperator": (
            "airflow.providers.google.cloud.operators.dataproc.DataprocJobBaseOperator"
        ),
        "DataprocScaleClusterOperator": (
            "airflow.providers.google.cloud.operators.dataproc.DataprocScaleClusterOperator"
        ),
        "DataprocSubmitHadoopJobOperator": (
            "airflow.providers.google.cloud.operators.dataproc.DataprocSubmitHadoopJobOperator"
        ),
        "DataprocSubmitHiveJobOperator": (
            "airflow.providers.google.cloud.operators.dataproc.DataprocSubmitHiveJobOperator"
        ),
        "DataprocSubmitPigJobOperator": (
            "airflow.providers.google.cloud.operators.dataproc.DataprocSubmitPigJobOperator"
        ),
        "DataprocSubmitPySparkJobOperator": (
            "airflow.providers.google.cloud.operators.dataproc.DataprocSubmitPySparkJobOperator"
        ),
        "DataprocSubmitSparkJobOperator": (
            "airflow.providers.google.cloud.operators.dataproc.DataprocSubmitSparkJobOperator"
        ),
        "DataprocSubmitSparkSqlJobOperator": (
            "airflow.providers.google.cloud.operators.dataproc.DataprocSubmitSparkSqlJobOperator"
        ),
        "DataprocClusterCreateOperator": (
            "airflow.providers.google.cloud.operators.dataproc.DataprocCreateClusterOperator"
        ),
        "DataprocClusterDeleteOperator": (
            "airflow.providers.google.cloud.operators.dataproc.DataprocDeleteClusterOperator"
        ),
        "DataprocClusterScaleOperator": (
            "airflow.providers.google.cloud.operators.dataproc.DataprocScaleClusterOperator"
        ),
        "DataProcHadoopOperator": (
            "airflow.providers.google.cloud.operators.dataproc.DataprocSubmitHadoopJobOperator"
        ),
        "DataProcHiveOperator": (
            "airflow.providers.google.cloud.operators.dataproc.DataprocSubmitHiveJobOperator"
        ),
        "DataProcJobBaseOperator": (
            "airflow.providers.google.cloud.operators.dataproc.DataprocJobBaseOperator"
        ),
        "DataProcPigOperator": (
            "airflow.providers.google.cloud.operators.dataproc.DataprocSubmitPigJobOperator"
        ),
        "DataProcPySparkOperator": (
            "airflow.providers.google.cloud.operators.dataproc.DataprocSubmitPySparkJobOperator"
        ),
        "DataProcSparkOperator": (
            "airflow.providers.google.cloud.operators.dataproc.DataprocSubmitSparkJobOperator"
        ),
        "DataProcSparkSqlOperator": (
            "airflow.providers.google.cloud.operators.dataproc.DataprocSubmitSparkSqlJobOperator"
        ),
        "DataprocWorkflowTemplateInstantiateInlineOperator":
            "airflow.providers.google.cloud.operators.dataproc."
            "DataprocInstantiateInlineWorkflowTemplateOperator",
        "DataprocWorkflowTemplateInstantiateOperator": (
            "airflow.providers.google.cloud.operators.dataproc.DataprocInstantiateWorkflowTemplateOperator"
        ),
    },
    "datastore_export_operator": {
        "CloudDatastoreExportEntitiesOperator": (
            "airflow.providers.google.cloud.operators.datastore.CloudDatastoreExportEntitiesOperator"
        ),
        "DatastoreExportOperator": (
            "airflow.providers.google.cloud.operators.datastore.CloudDatastoreExportEntitiesOperator"
        ),
    },
    "datastore_import_operator": {
        "CloudDatastoreImportEntitiesOperator": (
            "airflow.providers.google.cloud.operators.datastore.CloudDatastoreImportEntitiesOperator"
        ),
        "DatastoreImportOperator": (
            "airflow.providers.google.cloud.operators.datastore.CloudDatastoreImportEntitiesOperator"
        ),
    },
    "dingding_operator": {
        "DingdingOperator": "airflow.providers.dingding.operators.dingding.DingdingOperator",
    },
    "discord_webhook_operator": {
        "DiscordWebhookOperator": (
            "airflow.providers.discord.operators.discord_webhook.DiscordWebhookOperator"
        ),
    },
    "docker_swarm_operator": {
        "DockerSwarmOperator": "airflow.providers.docker.operators.docker_swarm.DockerSwarmOperator",
    },
    "druid_operator": {
        "DruidOperator": "airflow.providers.apache.druid.operators.druid.DruidOperator",
    },
    "dynamodb_to_s3": {
        "DynamoDBToS3Operator": "airflow.providers.amazon.aws.transfers.dynamodb_to_s3.DynamoDBToS3Operator",
    },
    "ecs_operator": {
        "EcsProtocol": "airflow.providers.amazon.aws.hooks.ecs.EcsProtocol",
        "EcsRunTaskOperator": "airflow.providers.amazon.aws.operators.ecs.EcsRunTaskOperator",
        "EcsOperator": "airflow.providers.amazon.aws.operators.ecs.EcsRunTaskOperator",
    },
    "file_to_gcs": {
        "LocalFilesystemToGCSOperator": (
            "airflow.providers.google.cloud.transfers.local_to_gcs.LocalFilesystemToGCSOperator"
        ),
        "FileToGoogleCloudStorageOperator": (
            "airflow.providers.google.cloud.transfers.local_to_gcs.LocalFilesystemToGCSOperator"
        ),
    },
    "file_to_wasb": {
        "LocalFilesystemToWasbOperator": (
            "airflow.providers.microsoft.azure.transfers.local_to_wasb.LocalFilesystemToWasbOperator"
        ),
        "FileToWasbOperator": (
            "airflow.providers.microsoft.azure.transfers.local_to_wasb.LocalFilesystemToWasbOperator"
        ),
    },
    "gcp_bigtable_operator": {
        "BigtableCreateInstanceOperator": (
            "airflow.providers.google.cloud.operators.bigtable.BigtableCreateInstanceOperator"
        ),
        "BigtableCreateTableOperator": (
            "airflow.providers.google.cloud.operators.bigtable.BigtableCreateTableOperator"
        ),
        "BigtableDeleteInstanceOperator": (
            "airflow.providers.google.cloud.operators.bigtable.BigtableDeleteInstanceOperator"
        ),
        "BigtableDeleteTableOperator": (
            "airflow.providers.google.cloud.operators.bigtable.BigtableDeleteTableOperator"
        ),
        "BigtableUpdateClusterOperator": (
            "airflow.providers.google.cloud.operators.bigtable.BigtableUpdateClusterOperator"
        ),
        "BigtableTableReplicationCompletedSensor": (
            "airflow.providers.google.cloud.sensors.bigtable.BigtableTableReplicationCompletedSensor"
        ),
        "BigtableClusterUpdateOperator": (
            "airflow.providers.google.cloud.operators.bigtable.BigtableUpdateClusterOperator"
        ),
        "BigtableInstanceCreateOperator": (
            "airflow.providers.google.cloud.operators.bigtable.BigtableCreateInstanceOperator"
        ),
        "BigtableInstanceDeleteOperator": (
            "airflow.providers.google.cloud.operators.bigtable.BigtableDeleteInstanceOperator"
        ),
        "BigtableTableCreateOperator": (
            "airflow.providers.google.cloud.operators.bigtable.BigtableCreateTableOperator"
        ),
        "BigtableTableDeleteOperator": (
            "airflow.providers.google.cloud.operators.bigtable.BigtableDeleteTableOperator"
        ),
        "BigtableTableWaitForReplicationSensor": (
            "airflow.providers.google.cloud.sensors.bigtable.BigtableTableReplicationCompletedSensor"
        ),
    },
    "gcp_cloud_build_operator": {
        "CloudBuildCreateBuildOperator": (
            "airflow.providers.google.cloud.operators.cloud_build.CloudBuildCreateBuildOperator"
        ),
    },
    "gcp_compute_operator": {
        "ComputeEngineBaseOperator": (
            "airflow.providers.google.cloud.operators.compute.ComputeEngineBaseOperator"
        ),
        "ComputeEngineCopyInstanceTemplateOperator": (
            "airflow.providers.google.cloud.operators.compute.ComputeEngineCopyInstanceTemplateOperator"
        ),
        "ComputeEngineInstanceGroupUpdateManagerTemplateOperator":
            "airflow.providers.google.cloud.operators.compute."
            "ComputeEngineInstanceGroupUpdateManagerTemplateOperator",
        "ComputeEngineSetMachineTypeOperator": (
            "airflow.providers.google.cloud.operators.compute.ComputeEngineSetMachineTypeOperator"
        ),
        "ComputeEngineStartInstanceOperator": (
            "airflow.providers.google.cloud.operators.compute.ComputeEngineStartInstanceOperator"
        ),
        "ComputeEngineStopInstanceOperator": (
            "airflow.providers.google.cloud.operators.compute.ComputeEngineStopInstanceOperator"
        ),
        "GceBaseOperator": "airflow.providers.google.cloud.operators.compute.ComputeEngineBaseOperator",
        "GceInstanceGroupManagerUpdateTemplateOperator":
            "airflow.providers.google.cloud.operators.compute."
            "ComputeEngineInstanceGroupUpdateManagerTemplateOperator",
        "GceInstanceStartOperator": (
            "airflow.providers.google.cloud.operators.compute.ComputeEngineStartInstanceOperator"
        ),
        "GceInstanceStopOperator": (
            "airflow.providers.google.cloud.operators.compute.ComputeEngineStopInstanceOperator"
        ),
        "GceInstanceTemplateCopyOperator": (
            "airflow.providers.google.cloud.operators.compute.ComputeEngineCopyInstanceTemplateOperator"
        ),
        "GceSetMachineTypeOperator": (
            "airflow.providers.google.cloud.operators.compute.ComputeEngineSetMachineTypeOperator"
        ),
    },
    "gcp_container_operator": {
        "GKECreateClusterOperator": (
            "airflow.providers.google.cloud.operators.kubernetes_engine.GKECreateClusterOperator"
        ),
        "GKEDeleteClusterOperator": (
            "airflow.providers.google.cloud.operators.kubernetes_engine.GKEDeleteClusterOperator"
        ),
        "GKEStartPodOperator": (
            "airflow.providers.google.cloud.operators.kubernetes_engine.GKEStartPodOperator"
        ),
        "GKEClusterCreateOperator": (
            "airflow.providers.google.cloud.operators.kubernetes_engine.GKECreateClusterOperator"
        ),
        "GKEClusterDeleteOperator": (
            "airflow.providers.google.cloud.operators.kubernetes_engine.GKEDeleteClusterOperator"
        ),
        "GKEPodOperator": "airflow.providers.google.cloud.operators.kubernetes_engine.GKEStartPodOperator",
    },
    "gcp_dlp_operator": {
        "CloudDLPCancelDLPJobOperator": (
            "airflow.providers.google.cloud.operators.dlp.CloudDLPCancelDLPJobOperator"
        ),
        "CloudDLPCreateDeidentifyTemplateOperator": (
            "airflow.providers.google.cloud.operators.dlp.CloudDLPCreateDeidentifyTemplateOperator"
        ),
        "CloudDLPCreateDLPJobOperator": (
            "airflow.providers.google.cloud.operators.dlp.CloudDLPCreateDLPJobOperator"
        ),
        "CloudDLPCreateInspectTemplateOperator": (
            "airflow.providers.google.cloud.operators.dlp.CloudDLPCreateInspectTemplateOperator"
        ),
        "CloudDLPCreateJobTriggerOperator": (
            "airflow.providers.google.cloud.operators.dlp.CloudDLPCreateJobTriggerOperator"
        ),
        "CloudDLPCreateStoredInfoTypeOperator": (
            "airflow.providers.google.cloud.operators.dlp.CloudDLPCreateStoredInfoTypeOperator"
        ),
        "CloudDLPDeidentifyContentOperator": (
            "airflow.providers.google.cloud.operators.dlp.CloudDLPDeidentifyContentOperator"
        ),
        "CloudDLPDeleteDeidentifyTemplateOperator": (
            "airflow.providers.google.cloud.operators.dlp.CloudDLPDeleteDeidentifyTemplateOperator"
        ),
        "CloudDLPDeleteDLPJobOperator": (
            "airflow.providers.google.cloud.operators.dlp.CloudDLPDeleteDLPJobOperator"
        ),
        "CloudDLPDeleteInspectTemplateOperator": (
            "airflow.providers.google.cloud.operators.dlp.CloudDLPDeleteInspectTemplateOperator"
        ),
        "CloudDLPDeleteJobTriggerOperator": (
            "airflow.providers.google.cloud.operators.dlp.CloudDLPDeleteJobTriggerOperator"
        ),
        "CloudDLPDeleteStoredInfoTypeOperator": (
            "airflow.providers.google.cloud.operators.dlp.CloudDLPDeleteStoredInfoTypeOperator"
        ),
        "CloudDLPGetDeidentifyTemplateOperator": (
            "airflow.providers.google.cloud.operators.dlp.CloudDLPGetDeidentifyTemplateOperator"
        ),
        "CloudDLPGetDLPJobOperator": "airflow.providers.google.cloud.operators.dlp.CloudDLPGetDLPJobOperator",
        "CloudDLPGetDLPJobTriggerOperator": (
            "airflow.providers.google.cloud.operators.dlp.CloudDLPGetDLPJobTriggerOperator"
        ),
        "CloudDLPGetInspectTemplateOperator": (
            "airflow.providers.google.cloud.operators.dlp.CloudDLPGetInspectTemplateOperator"
        ),
        "CloudDLPGetStoredInfoTypeOperator": (
            "airflow.providers.google.cloud.operators.dlp.CloudDLPGetStoredInfoTypeOperator"
        ),
        "CloudDLPInspectContentOperator": (
            "airflow.providers.google.cloud.operators.dlp.CloudDLPInspectContentOperator"
        ),
        "CloudDLPListDeidentifyTemplatesOperator": (
            "airflow.providers.google.cloud.operators.dlp.CloudDLPListDeidentifyTemplatesOperator"
        ),
        "CloudDLPListDLPJobsOperator": (
            "airflow.providers.google.cloud.operators.dlp.CloudDLPListDLPJobsOperator"
        ),
        "CloudDLPListInfoTypesOperator": (
            "airflow.providers.google.cloud.operators.dlp.CloudDLPListInfoTypesOperator"
        ),
        "CloudDLPListInspectTemplatesOperator": (
            "airflow.providers.google.cloud.operators.dlp.CloudDLPListInspectTemplatesOperator"
        ),
        "CloudDLPListJobTriggersOperator": (
            "airflow.providers.google.cloud.operators.dlp.CloudDLPListJobTriggersOperator"
        ),
        "CloudDLPListStoredInfoTypesOperator": (
            "airflow.providers.google.cloud.operators.dlp.CloudDLPListStoredInfoTypesOperator"
        ),
        "CloudDLPRedactImageOperator": (
            "airflow.providers.google.cloud.operators.dlp.CloudDLPRedactImageOperator"
        ),
        "CloudDLPReidentifyContentOperator": (
            "airflow.providers.google.cloud.operators.dlp.CloudDLPReidentifyContentOperator"
        ),
        "CloudDLPUpdateDeidentifyTemplateOperator": (
            "airflow.providers.google.cloud.operators.dlp.CloudDLPUpdateDeidentifyTemplateOperator"
        ),
        "CloudDLPUpdateInspectTemplateOperator": (
            "airflow.providers.google.cloud.operators.dlp.CloudDLPUpdateInspectTemplateOperator"
        ),
        "CloudDLPUpdateJobTriggerOperator": (
            "airflow.providers.google.cloud.operators.dlp.CloudDLPUpdateJobTriggerOperator"
        ),
        "CloudDLPUpdateStoredInfoTypeOperator": (
            "airflow.providers.google.cloud.operators.dlp.CloudDLPUpdateStoredInfoTypeOperator"
        ),
        "CloudDLPDeleteDlpJobOperator": (
            "airflow.providers.google.cloud.operators.dlp.CloudDLPDeleteDLPJobOperator"
        ),
        "CloudDLPGetDlpJobOperator": "airflow.providers.google.cloud.operators.dlp.CloudDLPGetDLPJobOperator",
        "CloudDLPGetJobTripperOperator": (
            "airflow.providers.google.cloud.operators.dlp.CloudDLPGetDLPJobTriggerOperator"
        ),
        "CloudDLPListDlpJobsOperator": (
            "airflow.providers.google.cloud.operators.dlp.CloudDLPListDLPJobsOperator"
        ),
    },
    "gcp_function_operator": {
        "CloudFunctionDeleteFunctionOperator": (
            "airflow.providers.google.cloud.operators.functions.CloudFunctionDeleteFunctionOperator"
        ),
        "CloudFunctionDeployFunctionOperator": (
            "airflow.providers.google.cloud.operators.functions.CloudFunctionDeployFunctionOperator"
        ),
        "GcfFunctionDeleteOperator": (
            "airflow.providers.google.cloud.operators.functions.CloudFunctionDeleteFunctionOperator"
        ),
        "GcfFunctionDeployOperator": (
            "airflow.providers.google.cloud.operators.functions.CloudFunctionDeployFunctionOperator"
        ),
    },
    "gcp_natural_language_operator": {
        "CloudNaturalLanguageAnalyzeEntitiesOperator":
            "airflow.providers.google.cloud.operators.natural_language."
            "CloudNaturalLanguageAnalyzeEntitiesOperator",
        "CloudNaturalLanguageAnalyzeEntitySentimentOperator":
            "airflow.providers.google.cloud.operators.natural_language."
            "CloudNaturalLanguageAnalyzeEntitySentimentOperator",
        "CloudNaturalLanguageAnalyzeSentimentOperator":
            "airflow.providers.google.cloud.operators.natural_language."
            "CloudNaturalLanguageAnalyzeSentimentOperator",
        "CloudNaturalLanguageClassifyTextOperator":
            "airflow.providers.google.cloud.operators.natural_language."
            "CloudNaturalLanguageClassifyTextOperator",
        "CloudLanguageAnalyzeEntitiesOperator":
            "airflow.providers.google.cloud.operators.natural_language."
            "CloudNaturalLanguageAnalyzeEntitiesOperator",
        "CloudLanguageAnalyzeEntitySentimentOperator":
            "airflow.providers.google.cloud.operators.natural_language."
            "CloudNaturalLanguageAnalyzeEntitySentimentOperator",
        "CloudLanguageAnalyzeSentimentOperator":
            "airflow.providers.google.cloud.operators.natural_language."
            "CloudNaturalLanguageAnalyzeSentimentOperator",
        "CloudLanguageClassifyTextOperator":
            "airflow.providers.google.cloud.operators.natural_language."
            "CloudNaturalLanguageClassifyTextOperator",
    },
    "gcp_spanner_operator": {
        "SpannerDeleteDatabaseInstanceOperator": (
            "airflow.providers.google.cloud.operators.spanner.SpannerDeleteDatabaseInstanceOperator"
        ),
        "SpannerDeleteInstanceOperator": (
            "airflow.providers.google.cloud.operators.spanner.SpannerDeleteInstanceOperator"
        ),
        "SpannerDeployDatabaseInstanceOperator": (
            "airflow.providers.google.cloud.operators.spanner.SpannerDeployDatabaseInstanceOperator"
        ),
        "SpannerDeployInstanceOperator": (
            "airflow.providers.google.cloud.operators.spanner.SpannerDeployInstanceOperator"
        ),
        "SpannerQueryDatabaseInstanceOperator": (
            "airflow.providers.google.cloud.operators.spanner.SpannerQueryDatabaseInstanceOperator"
        ),
        "SpannerUpdateDatabaseInstanceOperator": (
            "airflow.providers.google.cloud.operators.spanner.SpannerUpdateDatabaseInstanceOperator"
        ),
        "CloudSpannerInstanceDatabaseDeleteOperator": (
            "airflow.providers.google.cloud.operators.spanner.SpannerDeleteDatabaseInstanceOperator"
        ),
        "CloudSpannerInstanceDatabaseDeployOperator": (
            "airflow.providers.google.cloud.operators.spanner.SpannerDeployDatabaseInstanceOperator"
        ),
        "CloudSpannerInstanceDatabaseQueryOperator": (
            "airflow.providers.google.cloud.operators.spanner.SpannerQueryDatabaseInstanceOperator"
        ),
        "CloudSpannerInstanceDatabaseUpdateOperator": (
            "airflow.providers.google.cloud.operators.spanner.SpannerUpdateDatabaseInstanceOperator"
        ),
        "CloudSpannerInstanceDeleteOperator": (
            "airflow.providers.google.cloud.operators.spanner.SpannerDeleteInstanceOperator"
        ),
        "CloudSpannerInstanceDeployOperator": (
            "airflow.providers.google.cloud.operators.spanner.SpannerDeployInstanceOperator"
        ),
    },
    "gcp_speech_to_text_operator": {
        "CloudSpeechToTextRecognizeSpeechOperator": (
            "airflow.providers.google.cloud.operators.speech_to_text.CloudSpeechToTextRecognizeSpeechOperator"
        ),
        "GcpSpeechToTextRecognizeSpeechOperator": (
            "airflow.providers.google.cloud.operators.speech_to_text.CloudSpeechToTextRecognizeSpeechOperator"
        ),
    },
    "gcp_sql_operator": {
        "CloudSQLBaseOperator": "airflow.providers.google.cloud.operators.cloud_sql.CloudSQLBaseOperator",
        "CloudSQLCreateInstanceDatabaseOperator": (
            "airflow.providers.google.cloud.operators.cloud_sql.CloudSQLCreateInstanceDatabaseOperator"
        ),
        "CloudSQLCreateInstanceOperator": (
            "airflow.providers.google.cloud.operators.cloud_sql.CloudSQLCreateInstanceOperator"
        ),
        "CloudSQLDeleteInstanceDatabaseOperator": (
            "airflow.providers.google.cloud.operators.cloud_sql.CloudSQLDeleteInstanceDatabaseOperator"
        ),
        "CloudSQLDeleteInstanceOperator": (
            "airflow.providers.google.cloud.operators.cloud_sql.CloudSQLDeleteInstanceOperator"
        ),
        "CloudSQLExecuteQueryOperator": (
            "airflow.providers.google.cloud.operators.cloud_sql.CloudSQLExecuteQueryOperator"
        ),
        "CloudSQLExportInstanceOperator": (
            "airflow.providers.google.cloud.operators.cloud_sql.CloudSQLExportInstanceOperator"
        ),
        "CloudSQLImportInstanceOperator": (
            "airflow.providers.google.cloud.operators.cloud_sql.CloudSQLImportInstanceOperator"
        ),
        "CloudSQLInstancePatchOperator": (
            "airflow.providers.google.cloud.operators.cloud_sql.CloudSQLInstancePatchOperator"
        ),
        "CloudSQLPatchInstanceDatabaseOperator": (
            "airflow.providers.google.cloud.operators.cloud_sql.CloudSQLPatchInstanceDatabaseOperator"
        ),
        "CloudSqlBaseOperator": "airflow.providers.google.cloud.operators.cloud_sql.CloudSQLBaseOperator",
        "CloudSqlInstanceCreateOperator": (
            "airflow.providers.google.cloud.operators.cloud_sql.CloudSQLCreateInstanceOperator"
        ),
        "CloudSqlInstanceDatabaseCreateOperator": (
            "airflow.providers.google.cloud.operators.cloud_sql.CloudSQLCreateInstanceDatabaseOperator"
        ),
        "CloudSqlInstanceDatabaseDeleteOperator": (
            "airflow.providers.google.cloud.operators.cloud_sql.CloudSQLDeleteInstanceDatabaseOperator"
        ),
        "CloudSqlInstanceDatabasePatchOperator": (
            "airflow.providers.google.cloud.operators.cloud_sql.CloudSQLPatchInstanceDatabaseOperator"
        ),
        "CloudSqlInstanceDeleteOperator": (
            "airflow.providers.google.cloud.operators.cloud_sql.CloudSQLDeleteInstanceOperator"
        ),
        "CloudSqlInstanceExportOperator": (
            "airflow.providers.google.cloud.operators.cloud_sql.CloudSQLExportInstanceOperator"
        ),
        "CloudSqlInstanceImportOperator": (
            "airflow.providers.google.cloud.operators.cloud_sql.CloudSQLImportInstanceOperator"
        ),
        "CloudSqlInstancePatchOperator": (
            "airflow.providers.google.cloud.operators.cloud_sql.CloudSQLInstancePatchOperator"
        ),
        "CloudSqlQueryOperator": (
            "airflow.providers.google.cloud.operators.cloud_sql.CloudSQLExecuteQueryOperator"
        ),
    },
    "gcp_tasks_operator": {
        "CloudTasksQueueCreateOperator": (
            "airflow.providers.google.cloud.operators.tasks.CloudTasksQueueCreateOperator"
        ),
        "CloudTasksQueueDeleteOperator": (
            "airflow.providers.google.cloud.operators.tasks.CloudTasksQueueDeleteOperator"
        ),
        "CloudTasksQueueGetOperator": (
            "airflow.providers.google.cloud.operators.tasks.CloudTasksQueueGetOperator"
        ),
        "CloudTasksQueuePauseOperator": (
            "airflow.providers.google.cloud.operators.tasks.CloudTasksQueuePauseOperator"
        ),
        "CloudTasksQueuePurgeOperator": (
            "airflow.providers.google.cloud.operators.tasks.CloudTasksQueuePurgeOperator"
        ),
        "CloudTasksQueueResumeOperator": (
            "airflow.providers.google.cloud.operators.tasks.CloudTasksQueueResumeOperator"
        ),
        "CloudTasksQueuesListOperator": (
            "airflow.providers.google.cloud.operators.tasks.CloudTasksQueuesListOperator"
        ),
        "CloudTasksQueueUpdateOperator": (
            "airflow.providers.google.cloud.operators.tasks.CloudTasksQueueUpdateOperator"
        ),
        "CloudTasksTaskCreateOperator": (
            "airflow.providers.google.cloud.operators.tasks.CloudTasksTaskCreateOperator"
        ),
        "CloudTasksTaskDeleteOperator": (
            "airflow.providers.google.cloud.operators.tasks.CloudTasksTaskDeleteOperator"
        ),
        "CloudTasksTaskGetOperator": (
            "airflow.providers.google.cloud.operators.tasks.CloudTasksTaskGetOperator"
        ),
        "CloudTasksTaskRunOperator": (
            "airflow.providers.google.cloud.operators.tasks.CloudTasksTaskRunOperator"
        ),
        "CloudTasksTasksListOperator": (
            "airflow.providers.google.cloud.operators.tasks.CloudTasksTasksListOperator"
        ),
    },
    "gcp_text_to_speech_operator": {
        "CloudTextToSpeechSynthesizeOperator": (
            "airflow.providers.google.cloud.operators.text_to_speech.CloudTextToSpeechSynthesizeOperator"
        ),
        "GcpTextToSpeechSynthesizeOperator": (
            "airflow.providers.google.cloud.operators.text_to_speech.CloudTextToSpeechSynthesizeOperator"
        ),
    },
    "gcp_transfer_operator": {
        "CloudDataTransferServiceCancelOperationOperator":
            "airflow.providers.google.cloud.operators.cloud_storage_transfer_service."
            "CloudDataTransferServiceCancelOperationOperator",
        "CloudDataTransferServiceCreateJobOperator":
            "airflow.providers.google.cloud.operators.cloud_storage_transfer_service."
            "CloudDataTransferServiceCreateJobOperator",
        "CloudDataTransferServiceDeleteJobOperator":
            "airflow.providers.google.cloud.operators.cloud_storage_transfer_service."
            "CloudDataTransferServiceDeleteJobOperator",
        "CloudDataTransferServiceGCSToGCSOperator":
            "airflow.providers.google.cloud.operators.cloud_storage_transfer_service."
            "CloudDataTransferServiceGCSToGCSOperator",
        "CloudDataTransferServiceGetOperationOperator":
            "airflow.providers.google.cloud.operators.cloud_storage_transfer_service."
            "CloudDataTransferServiceGetOperationOperator",
        "CloudDataTransferServiceListOperationsOperator":
            "airflow.providers.google.cloud.operators.cloud_storage_transfer_service."
            "CloudDataTransferServiceListOperationsOperator",
        "CloudDataTransferServicePauseOperationOperator":
            "airflow.providers.google.cloud.operators.cloud_storage_transfer_service."
            "CloudDataTransferServicePauseOperationOperator",
        "CloudDataTransferServiceResumeOperationOperator":
            "airflow.providers.google.cloud.operators.cloud_storage_transfer_service."
            "CloudDataTransferServiceResumeOperationOperator",
        "CloudDataTransferServiceS3ToGCSOperator":
            "airflow.providers.google.cloud.operators.cloud_storage_transfer_service."
            "CloudDataTransferServiceS3ToGCSOperator",
        "CloudDataTransferServiceUpdateJobOperator":
            "airflow.providers.google.cloud.operators.cloud_storage_transfer_service."
            "CloudDataTransferServiceUpdateJobOperator",
        "GcpTransferServiceJobCreateOperator":
            "airflow.providers.google.cloud.operators.cloud_storage_transfer_service."
            "CloudDataTransferServiceCreateJobOperator",
        "GcpTransferServiceJobDeleteOperator":
            "airflow.providers.google.cloud.operators.cloud_storage_transfer_service."
            "CloudDataTransferServiceDeleteJobOperator",
        "GcpTransferServiceJobUpdateOperator":
            "airflow.providers.google.cloud.operators.cloud_storage_transfer_service."
            "CloudDataTransferServiceUpdateJobOperator",
        "GcpTransferServiceOperationCancelOperator":
            "airflow.providers.google.cloud.operators.cloud_storage_transfer_service."
            "CloudDataTransferServiceCancelOperationOperator",
        "GcpTransferServiceOperationGetOperator":
            "airflow.providers.google.cloud.operators.cloud_storage_transfer_service."
            "CloudDataTransferServiceGetOperationOperator",
        "GcpTransferServiceOperationPauseOperator":
            "airflow.providers.google.cloud.operators.cloud_storage_transfer_service."
            "CloudDataTransferServicePauseOperationOperator",
        "GcpTransferServiceOperationResumeOperator":
            "airflow.providers.google.cloud.operators.cloud_storage_transfer_service."
            "CloudDataTransferServiceResumeOperationOperator",
        "GcpTransferServiceOperationsListOperator":
            "airflow.providers.google.cloud.operators.cloud_storage_transfer_service."
            "CloudDataTransferServiceListOperationsOperator",
        "GoogleCloudStorageToGoogleCloudStorageTransferOperator":
            "airflow.providers.google.cloud.operators.cloud_storage_transfer_service."
            "CloudDataTransferServiceGCSToGCSOperator",
        "S3ToGoogleCloudStorageTransferOperator":
            "airflow.providers.google.cloud.operators.cloud_storage_transfer_service."
            "CloudDataTransferServiceS3ToGCSOperator",
    },
    "gcp_translate_operator": {
        "CloudTranslateTextOperator": (
            "airflow.providers.google.cloud.operators.translate.CloudTranslateTextOperator"
        ),
    },
    "gcp_translate_speech_operator": {
        "CloudTranslateSpeechOperator": (
            "airflow.providers.google.cloud.operators.translate_speech.CloudTranslateSpeechOperator"
        ),
        "GcpTranslateSpeechOperator": (
            "airflow.providers.google.cloud.operators.translate_speech.CloudTranslateSpeechOperator"
        ),
    },
    "gcp_video_intelligence_operator": {
        "CloudVideoIntelligenceDetectVideoExplicitContentOperator":
            "airflow.providers.google.cloud.operators.video_intelligence."
            "CloudVideoIntelligenceDetectVideoExplicitContentOperator",
        "CloudVideoIntelligenceDetectVideoLabelsOperator":
            "airflow.providers.google.cloud.operators.video_intelligence."
            "CloudVideoIntelligenceDetectVideoLabelsOperator",
        "CloudVideoIntelligenceDetectVideoShotsOperator":
            "airflow.providers.google.cloud.operators.video_intelligence."
            "CloudVideoIntelligenceDetectVideoShotsOperator",
    },
    "gcp_vision_operator": {
        "CloudVisionAddProductToProductSetOperator": (
            "airflow.providers.google.cloud.operators.vision.CloudVisionAddProductToProductSetOperator"
        ),
        "CloudVisionCreateProductOperator": (
            "airflow.providers.google.cloud.operators.vision.CloudVisionCreateProductOperator"
        ),
        "CloudVisionCreateProductSetOperator": (
            "airflow.providers.google.cloud.operators.vision.CloudVisionCreateProductSetOperator"
        ),
        "CloudVisionCreateReferenceImageOperator": (
            "airflow.providers.google.cloud.operators.vision.CloudVisionCreateReferenceImageOperator"
        ),
        "CloudVisionDeleteProductOperator": (
            "airflow.providers.google.cloud.operators.vision.CloudVisionDeleteProductOperator"
        ),
        "CloudVisionDeleteProductSetOperator": (
            "airflow.providers.google.cloud.operators.vision.CloudVisionDeleteProductSetOperator"
        ),
        "CloudVisionDetectImageLabelsOperator": (
            "airflow.providers.google.cloud.operators.vision.CloudVisionDetectImageLabelsOperator"
        ),
        "CloudVisionDetectImageSafeSearchOperator": (
            "airflow.providers.google.cloud.operators.vision.CloudVisionDetectImageSafeSearchOperator"
        ),
        "CloudVisionDetectTextOperator": (
            "airflow.providers.google.cloud.operators.vision.CloudVisionDetectTextOperator"
        ),
        "CloudVisionGetProductOperator": (
            "airflow.providers.google.cloud.operators.vision.CloudVisionGetProductOperator"
        ),
        "CloudVisionGetProductSetOperator": (
            "airflow.providers.google.cloud.operators.vision.CloudVisionGetProductSetOperator"
        ),
        "CloudVisionImageAnnotateOperator": (
            "airflow.providers.google.cloud.operators.vision.CloudVisionImageAnnotateOperator"
        ),
        "CloudVisionRemoveProductFromProductSetOperator": (
            "airflow.providers.google.cloud.operators.vision.CloudVisionRemoveProductFromProductSetOperator"
        ),
        "CloudVisionTextDetectOperator": (
            "airflow.providers.google.cloud.operators.vision.CloudVisionTextDetectOperator"
        ),
        "CloudVisionUpdateProductOperator": (
            "airflow.providers.google.cloud.operators.vision.CloudVisionUpdateProductOperator"
        ),
        "CloudVisionUpdateProductSetOperator": (
            "airflow.providers.google.cloud.operators.vision.CloudVisionUpdateProductSetOperator"
        ),
        "CloudVisionAnnotateImageOperator": (
            "airflow.providers.google.cloud.operators.vision.CloudVisionImageAnnotateOperator"
        ),
        "CloudVisionDetectDocumentTextOperator": (
            "airflow.providers.google.cloud.operators.vision.CloudVisionTextDetectOperator"
        ),
        "CloudVisionProductCreateOperator": (
            "airflow.providers.google.cloud.operators.vision.CloudVisionCreateProductOperator"
        ),
        "CloudVisionProductDeleteOperator": (
            "airflow.providers.google.cloud.operators.vision.CloudVisionDeleteProductOperator"
        ),
        "CloudVisionProductGetOperator": (
            "airflow.providers.google.cloud.operators.vision.CloudVisionGetProductOperator"
        ),
        "CloudVisionProductSetCreateOperator": (
            "airflow.providers.google.cloud.operators.vision.CloudVisionCreateProductSetOperator"
        ),
        "CloudVisionProductSetDeleteOperator": (
            "airflow.providers.google.cloud.operators.vision.CloudVisionDeleteProductSetOperator"
        ),
        "CloudVisionProductSetGetOperator": (
            "airflow.providers.google.cloud.operators.vision.CloudVisionGetProductSetOperator"
        ),
        "CloudVisionProductSetUpdateOperator": (
            "airflow.providers.google.cloud.operators.vision.CloudVisionUpdateProductSetOperator"
        ),
        "CloudVisionProductUpdateOperator": (
            "airflow.providers.google.cloud.operators.vision.CloudVisionUpdateProductOperator"
        ),
        "CloudVisionReferenceImageCreateOperator": (
            "airflow.providers.google.cloud.operators.vision.CloudVisionCreateReferenceImageOperator"
        ),
    },
    "gcs_acl_operator": {
        "GCSBucketCreateAclEntryOperator": (
            "airflow.providers.google.cloud.operators.gcs.GCSBucketCreateAclEntryOperator"
        ),
        "GCSObjectCreateAclEntryOperator": (
            "airflow.providers.google.cloud.operators.gcs.GCSObjectCreateAclEntryOperator"
        ),
        "GoogleCloudStorageBucketCreateAclEntryOperator": (
            "airflow.providers.google.cloud.operators.gcs.GCSBucketCreateAclEntryOperator"
        ),
        "GoogleCloudStorageObjectCreateAclEntryOperator": (
            "airflow.providers.google.cloud.operators.gcs.GCSObjectCreateAclEntryOperator"
        ),
    },
    "gcs_delete_operator": {
        "GCSDeleteObjectsOperator": "airflow.providers.google.cloud.operators.gcs.GCSDeleteObjectsOperator",
        "GoogleCloudStorageDeleteOperator": (
            "airflow.providers.google.cloud.operators.gcs.GCSDeleteObjectsOperator"
        ),
    },
    "gcs_download_operator": {
        "GCSToLocalFilesystemOperator": (
            "airflow.providers.google.cloud.transfers.gcs_to_local.GCSToLocalFilesystemOperator"
        ),
        "GoogleCloudStorageDownloadOperator": (
            "airflow.providers.google.cloud.transfers.gcs_to_local.GCSToLocalFilesystemOperator"
        ),
    },
    "gcs_list_operator": {
        "GCSListObjectsOperator": "airflow.providers.google.cloud.operators.gcs.GCSListObjectsOperator",
        "GoogleCloudStorageListOperator": (
            "airflow.providers.google.cloud.operators.gcs.GCSListObjectsOperator"
        ),
    },
    "gcs_operator": {
        "GCSCreateBucketOperator": "airflow.providers.google.cloud.operators.gcs.GCSCreateBucketOperator",
        "GoogleCloudStorageCreateBucketOperator": (
            "airflow.providers.google.cloud.operators.gcs.GCSCreateBucketOperator"
        ),
    },
    "gcs_to_bq": {
        "GCSToBigQueryOperator": (
            "airflow.providers.google.cloud.transfers.gcs_to_bigquery.GCSToBigQueryOperator"
        ),
        "GoogleCloudStorageToBigQueryOperator": (
            "airflow.providers.google.cloud.transfers.gcs_to_bigquery.GCSToBigQueryOperator"
        ),
    },
    "gcs_to_gcs": {
        "GCSToGCSOperator": "airflow.providers.google.cloud.transfers.gcs_to_gcs.GCSToGCSOperator",
        "GoogleCloudStorageToGoogleCloudStorageOperator": (
            "airflow.providers.google.cloud.transfers.gcs_to_gcs.GCSToGCSOperator"
        ),
    },
    "gcs_to_gdrive_operator": {
        "GCSToGoogleDriveOperator": (
            "airflow.providers.google.suite.transfers.gcs_to_gdrive.GCSToGoogleDriveOperator"
        ),
    },
    "gcs_to_s3": {
        "GCSToS3Operator": "airflow.providers.amazon.aws.transfers.gcs_to_s3.GCSToS3Operator",
        "GoogleCloudStorageToS3Operator": "airflow.providers.amazon.aws.transfers.gcs_to_s3.GCSToS3Operator",
    },
    "grpc_operator": {
        "GrpcOperator": "airflow.providers.grpc.operators.grpc.GrpcOperator",
    },
    "hive_to_dynamodb": {
        "HiveToDynamoDBOperator": (
            "airflow.providers.amazon.aws.transfers.hive_to_dynamodb.HiveToDynamoDBOperator"
        ),
    },
    "imap_attachment_to_s3_operator": {
        "ImapAttachmentToS3Operator": (
            "airflow.providers.amazon.aws.transfers.imap_attachment_to_s3.ImapAttachmentToS3Operator"
        ),
    },
    "jenkins_job_trigger_operator": {
        "JenkinsJobTriggerOperator": (
            "airflow.providers.jenkins.operators.jenkins_job_trigger.JenkinsJobTriggerOperator"
        ),
    },
    "jira_operator": {
        "JiraOperator": "airflow.providers.atlassian.jira.operators.jira.JiraOperator",
    },
    "kubernetes_pod_operator": {
        "KubernetesPodOperator": (
            "airflow.providers.cncf.kubernetes.operators.pod.KubernetesPodOperator"
        ),
    },
    "mlengine_operator": {
        "MLEngineManageModelOperator": (
            "airflow.providers.google.cloud.operators.mlengine.MLEngineManageModelOperator"
        ),
        "MLEngineManageVersionOperator": (
            "airflow.providers.google.cloud.operators.mlengine.MLEngineManageVersionOperator"
        ),
        "MLEngineStartBatchPredictionJobOperator": (
            "airflow.providers.google.cloud.operators.mlengine.MLEngineStartBatchPredictionJobOperator"
        ),
        "MLEngineStartTrainingJobOperator": (
            "airflow.providers.google.cloud.operators.mlengine.MLEngineStartTrainingJobOperator"
        ),
        "MLEngineBatchPredictionOperator": (
            "airflow.providers.google.cloud.operators.mlengine.MLEngineStartBatchPredictionJobOperator"
        ),
        "MLEngineModelOperator": (
            "airflow.providers.google.cloud.operators.mlengine.MLEngineManageModelOperator"
        ),
        "MLEngineTrainingOperator": (
            "airflow.providers.google.cloud.operators.mlengine.MLEngineStartTrainingJobOperator"
        ),
        "MLEngineVersionOperator": (
            "airflow.providers.google.cloud.operators.mlengine.MLEngineManageVersionOperator"
        ),
    },
    "mongo_to_s3": {
        "MongoToS3Operator": "airflow.providers.amazon.aws.transfers.mongo_to_s3.MongoToS3Operator",
    },
    "mssql_to_gcs": {
        "MSSQLToGCSOperator": "airflow.providers.google.cloud.transfers.mssql_to_gcs.MSSQLToGCSOperator",
        "MsSqlToGoogleCloudStorageOperator": (
            "airflow.providers.google.cloud.transfers.mssql_to_gcs.MSSQLToGCSOperator"
        ),
    },
    "mysql_to_gcs": {
        "MySQLToGCSOperator": "airflow.providers.google.cloud.transfers.mysql_to_gcs.MySQLToGCSOperator",
        "MySqlToGoogleCloudStorageOperator": (
            "airflow.providers.google.cloud.transfers.mysql_to_gcs.MySQLToGCSOperator"
        ),
    },
    "opsgenie_alert_operator": {
        "OpsgenieCreateAlertOperator": (
            "airflow.providers.opsgenie.operators.opsgenie.OpsgenieCreateAlertOperator"
        ),
        "OpsgenieAlertOperator": "airflow.providers.opsgenie.operators.opsgenie.OpsgenieCreateAlertOperator",
    },
    "oracle_to_azure_data_lake_transfer": {
        "OracleToAzureDataLakeOperator":
            "airflow.providers.microsoft.azure.transfers."
            "oracle_to_azure_data_lake.OracleToAzureDataLakeOperator",
    },
    "oracle_to_oracle_transfer": {
        "OracleToOracleOperator": (
            "airflow.providers.oracle.transfers.oracle_to_oracle.OracleToOracleOperator"
        ),
        "OracleToOracleTransfer": (
            "airflow.providers.oracle.transfers.oracle_to_oracle.OracleToOracleOperator"
        ),
    },
    "postgres_to_gcs_operator": {
        "PostgresToGCSOperator": (
            "airflow.providers.google.cloud.transfers.postgres_to_gcs.PostgresToGCSOperator"
        ),
        "PostgresToGoogleCloudStorageOperator": (
            "airflow.providers.google.cloud.transfers.postgres_to_gcs.PostgresToGCSOperator"
        ),
    },
    "pubsub_operator": {
        "PubSubCreateSubscriptionOperator": (
            "airflow.providers.google.cloud.operators.pubsub.PubSubCreateSubscriptionOperator"
        ),
        "PubSubCreateTopicOperator": (
            "airflow.providers.google.cloud.operators.pubsub.PubSubCreateTopicOperator"
        ),
        "PubSubDeleteSubscriptionOperator": (
            "airflow.providers.google.cloud.operators.pubsub.PubSubDeleteSubscriptionOperator"
        ),
        "PubSubDeleteTopicOperator": (
            "airflow.providers.google.cloud.operators.pubsub.PubSubDeleteTopicOperator"
        ),
        "PubSubPublishMessageOperator": (
            "airflow.providers.google.cloud.operators.pubsub.PubSubPublishMessageOperator"
        ),
        "PubSubPublishOperator": (
            "airflow.providers.google.cloud.operators.pubsub.PubSubPublishMessageOperator"
        ),
        "PubSubSubscriptionCreateOperator": (
            "airflow.providers.google.cloud.operators.pubsub.PubSubCreateSubscriptionOperator"
        ),
        "PubSubSubscriptionDeleteOperator": (
            "airflow.providers.google.cloud.operators.pubsub.PubSubDeleteSubscriptionOperator"
        ),
        "PubSubTopicCreateOperator": (
            "airflow.providers.google.cloud.operators.pubsub.PubSubCreateTopicOperator"
        ),
        "PubSubTopicDeleteOperator": (
            "airflow.providers.google.cloud.operators.pubsub.PubSubDeleteTopicOperator"
        ),
    },
    "redis_publish_operator": {
        "RedisPublishOperator": "airflow.providers.redis.operators.redis_publish.RedisPublishOperator",
    },
    "s3_to_gcs_operator": {
        "S3ToGCSOperator": "airflow.providers.google.cloud.transfers.s3_to_gcs.S3ToGCSOperator",
    },
    "s3_to_gcs_transfer_operator": {
        "CloudDataTransferServiceS3ToGCSOperator":
            "airflow.providers.google.cloud.operators.cloud_storage_transfer_service."
            "CloudDataTransferServiceS3ToGCSOperator",
    },
    "s3_to_sftp_operator": {
        "S3ToSFTPOperator": "airflow.providers.amazon.aws.transfers.s3_to_sftp.S3ToSFTPOperator",
    },
    "segment_track_event_operator": {
        "SegmentTrackEventOperator": (
            "airflow.providers.segment.operators.segment_track_event.SegmentTrackEventOperator"
        ),
    },
    "sftp_operator": {
        "SFTPOperator": "airflow.providers.sftp.operators.sftp.SFTPOperator",
    },
    "sftp_to_s3_operator": {
        "SFTPToS3Operator": "airflow.providers.amazon.aws.transfers.sftp_to_s3.SFTPToS3Operator",
    },
    "slack_webhook_operator": {
        "SlackWebhookOperator": "airflow.providers.slack.operators.slack_webhook.SlackWebhookOperator",
    },
    "snowflake_operator": {
        "SnowflakeOperator": "airflow.providers.snowflake.operators.snowflake.SnowflakeOperator",
    },
    "sns_publish_operator": {
        "SnsPublishOperator": "airflow.providers.amazon.aws.operators.sns.SnsPublishOperator",
    },
    "spark_jdbc_operator": {
        "SparkJDBCOperator": "airflow.providers.apache.spark.operators.spark_jdbc.SparkJDBCOperator",
        "SparkSubmitOperator": "airflow.providers.apache.spark.operators.spark_jdbc.SparkSubmitOperator",
    },
    "spark_sql_operator": {
        "SparkSqlOperator": "airflow.providers.apache.spark.operators.spark_sql.SparkSqlOperator",
    },
    "spark_submit_operator": {
        "SparkSubmitOperator": "airflow.providers.apache.spark.operators.spark_submit.SparkSubmitOperator",
    },
    "sql_to_gcs": {
        "BaseSQLToGCSOperator": "airflow.providers.google.cloud.transfers.sql_to_gcs.BaseSQLToGCSOperator",
        "BaseSQLToGoogleCloudStorageOperator": (
            "airflow.providers.google.cloud.transfers.sql_to_gcs.BaseSQLToGCSOperator"
        ),
    },
    "sqoop_operator": {
        "SqoopOperator": "airflow.providers.apache.sqoop.operators.sqoop.SqoopOperator",
    },
    "ssh_operator": {
        "SSHOperator": "airflow.providers.ssh.operators.ssh.SSHOperator",
    },
    "vertica_operator": {
        "VerticaOperator": "airflow.providers.vertica.operators.vertica.VerticaOperator",
    },
    "vertica_to_hive": {
        "VerticaToHiveOperator": (
            "airflow.providers.apache.hive.transfers.vertica_to_hive.VerticaToHiveOperator"
        ),
        "VerticaToHiveTransfer": (
            "airflow.providers.apache.hive.transfers.vertica_to_hive.VerticaToHiveOperator"
        ),
    },
    "vertica_to_mysql": {
        "VerticaToMySqlOperator": "airflow.providers.mysql.transfers.vertica_to_mysql.VerticaToMySqlOperator",
        "VerticaToMySqlTransfer": "airflow.providers.mysql.transfers.vertica_to_mysql.VerticaToMySqlOperator",
    },
    "wasb_delete_blob_operator": {
        "WasbDeleteBlobOperator": (
            "airflow.providers.microsoft.azure.operators.wasb_delete_blob.WasbDeleteBlobOperator"
        ),
    },
    "winrm_operator": {
        "WinRMOperator": "airflow.providers.microsoft.winrm.operators.winrm.WinRMOperator",
    },
}

add_deprecated_classes(__deprecated_classes, __name__)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import contextlib
from functools import wraps
from inspect import signature
from typing import Callable, Generator, TypeVar, cast

from sqlalchemy.orm import Session as SASession

from airflow import settings
from airflow.typing_compat import ParamSpec


@contextlib.contextmanager
def create_session() -> Generator[SASession, None, None]:
    """Contextmanager that will create and teardown a session."""
    Session = getattr(settings, "Session", None)
    if Session is None:
        raise RuntimeError("Session must be set before!")
    session = Session()
    try:
        yield session
        session.commit()
    except Exception:
        session.rollback()
        raise
    finally:
        session.close()


PS = ParamSpec("PS")
RT = TypeVar("RT")


def find_session_idx(func: Callable[PS, RT]) -> int:
    """Find session index in function call parameter."""
    func_params = signature(func).parameters
    try:
        # func_params is an ordered dict -- this is the "recommended" way of getting the position
        session_args_idx = tuple(func_params).index("session")
    except ValueError:
        raise ValueError(f"Function {func.__qualname__} has no `session` argument") from None

    return session_args_idx


def provide_session(func: Callable[PS, RT]) -> Callable[PS, RT]:
    """
    Provide a session if it isn't provided.

    If you want to reuse a session or run the function as part of a
    database transaction, you pass it to the function, if not this wrapper
    will create one and close it for you.
    """
    session_args_idx = find_session_idx(func)

    @wraps(func)
    def wrapper(*args, **kwargs) -> RT:
        if "session" in kwargs or session_args_idx < len(args):
            return func(*args, **kwargs)
        else:
            with create_session() as session:
                return func(*args, session=session, **kwargs)

    return wrapper


# A fake session to use in functions decorated by provide_session. This allows
# the 'session' argument to be of type Session instead of Session | None,
# making it easier to type hint the function body without dealing with the None
# case that can never happen at runtime.
NEW_SESSION: SASession = cast(SASession, None)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

# [START tutorial]
# [START import_module]
import pendulum
import requests

from airflow.decorators import dag, task
from airflow.io.path import ObjectStoragePath

# [END import_module]

API = "https://opendata.fmi.fi/timeseries"

aq_fields = {
    "fmisid": "int32",
    "time": "datetime64[ns]",
    "AQINDEX_PT1H_avg": "float64",
    "PM10_PT1H_avg": "float64",
    "PM25_PT1H_avg": "float64",
    "O3_PT1H_avg": "float64",
    "CO_PT1H_avg": "float64",
    "SO2_PT1H_avg": "float64",
    "NO2_PT1H_avg": "float64",
    "TRSC_PT1H_avg": "float64",
}

# [START create_object_storage_path]
base = ObjectStoragePath("s3://aws_default@airflow-tutorial-data/")
# [END create_object_storage_path]


@dag(
    schedule=None,
    start_date=pendulum.datetime(2021, 1, 1, tz="UTC"),
    catchup=False,
    tags=["example"],
)
def tutorial_objectstorage():
    """
    ### Object Storage Tutorial Documentation
    This is a tutorial DAG to showcase the usage of the Object Storage API.
    Documentation that goes along with the Airflow Object Storage tutorial is
    located
    [here](https://airflow.apache.org/docs/apache-airflow/stable/tutorial/objectstorage.html)
    """

    # [START get_air_quality_data]
    @task
    def get_air_quality_data(**kwargs) -> ObjectStoragePath:
        """
        #### Get Air Quality Data
        This task gets air quality data from the Finnish Meteorological Institute's
        open data API. The data is saved as parquet.
        """
        import pandas as pd

        execution_date = kwargs["logical_date"]
        start_time = kwargs["data_interval_start"]

        params = {
            "format": "json",
            "precision": "double",
            "groupareas": "0",
            "producer": "airquality_urban",
            "area": "Uusimaa",
            "param": ",".join(aq_fields.keys()),
            "starttime": start_time.isoformat(timespec="seconds"),
            "endtime": execution_date.isoformat(timespec="seconds"),
            "tz": "UTC",
        }

        response = requests.get(API, params=params)
        response.raise_for_status()

        # ensure the bucket exists
        base.mkdir(exist_ok=True)

        formatted_date = execution_date.format("YYYYMMDD")
        path = base / f"air_quality_{formatted_date}.parquet"

        df = pd.DataFrame(response.json()).astype(aq_fields)
        with path.open("wb") as file:
            df.to_parquet(file)

        return path

    # [END get_air_quality_data]

    # [START analyze]
    @task
    def analyze(path: ObjectStoragePath, **kwargs):
        """
        #### Analyze
        This task analyzes the air quality data, prints the results
        """
        import duckdb

        conn = duckdb.connect(database=":memory:")
        conn.register_filesystem(path.fs)
        conn.execute(f"CREATE OR REPLACE TABLE airquality_urban AS SELECT * FROM read_parquet('{path}')")

        df2 = conn.execute("SELECT * FROM airquality_urban").fetchdf()

        print(df2.head())

    # [END analyze]

    # [START main_flow]
    obj_path = get_air_quality_data()
    analyze(obj_path)
    # [END main_flow]


# [START dag_invocation]
tutorial_objectstorage()
# [END dag_invocation]
# [END tutorial]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING, ClassVar

import attr

if TYPE_CHECKING:
    from airflow.models.baseoperator import BaseOperator
    from airflow.models.taskinstancekey import TaskInstanceKey


@attr.s(auto_attribs=True)
class BaseOperatorLink(metaclass=ABCMeta):
    """Abstract base class that defines how we get an operator link."""

    operators: ClassVar[list[type[BaseOperator]]] = []
    """
    This property will be used by Airflow Plugins to find the Operators to which you want
    to assign this Operator Link

    :return: List of Operator classes used by task for which you want to create extra link
    """

    @property
    @abstractmethod
    def name(self) -> str:
        """Name of the link. This will be the button name on the task UI."""

    @abstractmethod
    def get_link(self, operator: BaseOperator, *, ti_key: TaskInstanceKey) -> str:
        """Link to external system.

        Note: The old signature of this function was ``(self, operator, dttm: datetime)``. That is still
        supported at runtime but is deprecated.

        :param operator: The Airflow operator object this link is associated to.
        :param ti_key: TaskInstance ID to return link for.
        :return: link to external system
        """



#!/usr/bin/env python
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Renderer DAG (tasks and dependencies) to the graphviz object."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any

import graphviz

from airflow.exceptions import AirflowException
from airflow.models.baseoperator import BaseOperator
from airflow.models.mappedoperator import MappedOperator
from airflow.utils.dag_edges import dag_edges
from airflow.utils.state import State
from airflow.utils.task_group import TaskGroup

if TYPE_CHECKING:
    from airflow.models import TaskInstance
    from airflow.models.dag import DAG
    from airflow.models.taskmixin import DependencyMixin
    from airflow.serialization.serialized_objects import DagDependency


def _refine_color(color: str):
    """
    Convert color in #RGB (12 bits) format to #RRGGBB (32 bits), if it possible.

    Otherwise, it returns the original value. Graphviz does not support colors in #RGB format.

    :param color: Text representation of color
    :return: Refined representation of color
    """
    if len(color) == 4 and color[0] == "#":
        color_r = color[1]
        color_g = color[2]
        color_b = color[3]
        return "#" + color_r + color_r + color_g + color_g + color_b + color_b
    return color


def _draw_task(
    task: MappedOperator | BaseOperator,
    parent_graph: graphviz.Digraph,
    states_by_task_id: dict[Any, Any] | None,
) -> None:
    """Draw a single task on the given parent_graph."""
    if states_by_task_id:
        state = states_by_task_id.get(task.task_id)
        color = State.color_fg(state)
        fill_color = State.color(state)
    else:
        color = task.ui_fgcolor
        fill_color = task.ui_color

    parent_graph.node(
        task.task_id,
        _attributes={
            "label": task.label,
            "shape": "rectangle",
            "style": "filled,rounded",
            "color": _refine_color(color),
            "fillcolor": _refine_color(fill_color),
        },
    )


def _draw_task_group(
    task_group: TaskGroup, parent_graph: graphviz.Digraph, states_by_task_id: dict[str, str] | None
) -> None:
    """Draw the given task_group and its children on the given parent_graph."""
    # Draw joins
    if task_group.upstream_group_ids or task_group.upstream_task_ids:
        parent_graph.node(
            task_group.upstream_join_id,
            _attributes={
                "label": "",
                "shape": "circle",
                "style": "filled,rounded",
                "color": _refine_color(task_group.ui_fgcolor),
                "fillcolor": _refine_color(task_group.ui_color),
                "width": "0.2",
                "height": "0.2",
            },
        )

    if task_group.downstream_group_ids or task_group.downstream_task_ids:
        parent_graph.node(
            task_group.downstream_join_id,
            _attributes={
                "label": "",
                "shape": "circle",
                "style": "filled,rounded",
                "color": _refine_color(task_group.ui_fgcolor),
                "fillcolor": _refine_color(task_group.ui_color),
                "width": "0.2",
                "height": "0.2",
            },
        )

    # Draw children
    for child in sorted(task_group.children.values(), key=lambda t: t.node_id if t.node_id else ""):
        _draw_nodes(child, parent_graph, states_by_task_id)


def _draw_nodes(
    node: DependencyMixin, parent_graph: graphviz.Digraph, states_by_task_id: dict[str, str] | None
) -> None:
    """Draw the node and its children on the given parent_graph recursively."""
    if isinstance(node, (BaseOperator, MappedOperator)):
        _draw_task(node, parent_graph, states_by_task_id)
    else:
        if not isinstance(node, TaskGroup):
            raise AirflowException(f"The node {node} should be TaskGroup and is not")
        # Draw TaskGroup
        if node.is_root:
            # No need to draw background for root TaskGroup.
            _draw_task_group(node, parent_graph, states_by_task_id)
        else:
            with parent_graph.subgraph(name=f"cluster_{node.group_id}") as sub:
                sub.attr(
                    shape="rectangle",
                    style="filled",
                    color=_refine_color(node.ui_fgcolor),
                    # Partially transparent CornflowerBlue
                    fillcolor="#6495ed7f",
                    label=node.label,
                )
                _draw_task_group(node, sub, states_by_task_id)


def render_dag_dependencies(deps: dict[str, list[DagDependency]]) -> graphviz.Digraph:
    """
    Render the DAG dependency to the DOT object.

    :param deps: List of DAG dependencies
    :return: Graphviz object
    """
    dot = graphviz.Digraph(graph_attr={"rankdir": "LR"})

    for dag, dependencies in deps.items():
        for dep in dependencies:
            with dot.subgraph(
                name=dag,
                graph_attr={
                    "rankdir": "LR",
                    "labelloc": "t",
                    "label": dag,
                },
            ) as dep_subgraph:
                dep_subgraph.edge(dep.source, dep.dependency_id)
                dep_subgraph.edge(dep.dependency_id, dep.target)

    return dot


def render_dag(dag: DAG, tis: list[TaskInstance] | None = None) -> graphviz.Digraph:
    """
    Render the DAG object to the DOT object.

    If an task instance list is passed, the nodes will be painted according to task statuses.

    :param dag: DAG that will be rendered.
    :param tis: List of task instances
    :return: Graphviz object
    """
    dot = graphviz.Digraph(
        dag.dag_id,
        graph_attr={
            "rankdir": dag.orientation if dag.orientation else "LR",
            "labelloc": "t",
            "label": dag.dag_id,
        },
    )
    states_by_task_id = None
    if tis is not None:
        states_by_task_id = {ti.task_id: ti.state for ti in tis}

    _draw_nodes(dag.task_group, dot, states_by_task_id)

    for edge in dag_edges(dag):
        # Gets an optional label for the edge; this will be None if none is specified.
        label = dag.get_edge_info(edge["source_id"], edge["target_id"]).get("label")
        # Add the edge to the graph with optional label
        # (we can just use the maybe-None label variable directly)
        dot.edge(edge["source_id"], edge["target_id"], label)

    return dot



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import io
import sys
from typing import TYPE_CHECKING, Collection

if TYPE_CHECKING:
    from io import IOBase


class CliConflictError(Exception):
    """Error for when CLI commands are defined twice by different sources."""

    pass


def is_stdout(fileio: IOBase) -> bool:
    """Check whether a file IO is stdout.

    The intended use case for this helper is to check whether an argument parsed
    with argparse.FileType points to stdout (by setting the path to ``-``). This
    is why there is no equivalent for stderr; argparse does not allow using it.

    .. warning:: *fileio* must be open for this check to be successful.
    """
    return fileio.fileno() == sys.stdout.fileno()


def print_export_output(command_type: str, exported_items: Collection, file: io.TextIOWrapper):
    if not file.closed and is_stdout(file):
        print(f"\n{len(exported_items)} {command_type} successfully exported.", file=sys.stderr)
    else:
        print(f"{len(exported_items)} {command_type} successfully exported to {file.name}.")



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING

from airflow.utils.module_loading import qualname

serializers = ["pyiceberg.table.Table"]
deserializers = serializers
stringifiers = serializers

if TYPE_CHECKING:
    from airflow.serialization.serde import U

__version__ = 1


def serialize(o: object) -> tuple[U, str, int, bool]:
    from pyiceberg.table import Table

    if not isinstance(o, Table):
        return "", "", 0, False

    from airflow.models.crypto import get_fernet

    # we encrypt the catalog information here until we have
    # global catalog management in airflow and the properties
    # can have sensitive information
    fernet = get_fernet()
    properties = {}
    for k, v in o.catalog.properties.items():
        properties[k] = fernet.encrypt(v.encode("utf-8")).decode("utf-8")

    data = {
        "identifier": o.identifier,
        "catalog_properties": properties,
    }

    return data, qualname(o), __version__, True


def deserialize(classname: str, version: int, data: dict):
    from pyiceberg.catalog import load_catalog
    from pyiceberg.table import Table

    from airflow.models.crypto import get_fernet

    if version > __version__:
        raise TypeError("serialized version is newer than class version")

    if classname == qualname(Table):
        fernet = get_fernet()
        properties = {}
        for k, v in data["catalog_properties"].items():
            properties[k] = fernet.decrypt(v.encode("utf-8")).decode("utf-8")

        catalog = load_catalog(data["identifier"][0], **properties)
        return catalog.load_table((data["identifier"][1], data["identifier"][2]))

    raise TypeError(f"do not know how to deserialize {classname}")



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
# fmt: off
"""Operators."""
from __future__ import annotations

from airflow.utils.deprecation_tools import add_deprecated_classes

__deprecated_classes = {
    "bash_operator": {
        "BashOperator": "airflow.operators.bash.BashOperator",
    },
    "branch_operator": {
        "BaseBranchOperator": "airflow.operators.branch.BaseBranchOperator",
    },
    "check_operator": {
        "SQLCheckOperator": "airflow.providers.common.sql.operators.sql.SQLCheckOperator",
        "SQLIntervalCheckOperator": "airflow.providers.common.sql.operators.sql.SQLIntervalCheckOperator",
        "SQLThresholdCheckOperator": "airflow.providers.common.sql.operators.sql.SQLThresholdCheckOperator",
        "SQLValueCheckOperator": "airflow.providers.common.sql.operators.sql.SQLValueCheckOperator",
        "CheckOperator": "airflow.providers.common.sql.operators.sql.SQLCheckOperator",
        "IntervalCheckOperator": "airflow.providers.common.sql.operators.sql.SQLIntervalCheckOperator",
        "ThresholdCheckOperator": "airflow.providers.common.sql.operators.sql.SQLThresholdCheckOperator",
        "ValueCheckOperator": "airflow.providers.common.sql.operators.sql.SQLValueCheckOperator",
    },
    "dagrun_operator": {
        "TriggerDagRunLink": "airflow.operators.trigger_dagrun.TriggerDagRunLink",
        "TriggerDagRunOperator": "airflow.operators.trigger_dagrun.TriggerDagRunOperator",
    },
    "docker_operator": {
        "DockerOperator": "airflow.providers.docker.operators.docker.DockerOperator",
    },
    "druid_check_operator": {
        "DruidCheckOperator": "airflow.providers.apache.druid.operators.druid_check.DruidCheckOperator",
    },
    "dummy": {
        "EmptyOperator": "airflow.operators.empty.EmptyOperator",
        "DummyOperator": "airflow.operators.empty.EmptyOperator",
    },
    "dummy_operator": {
        "EmptyOperator": "airflow.operators.empty.EmptyOperator",
        "DummyOperator": "airflow.operators.empty.EmptyOperator",
    },
    "email_operator": {
        "EmailOperator": "airflow.operators.email.EmailOperator",
    },
    "gcs_to_s3": {
        "GCSToS3Operator": "airflow.providers.amazon.aws.transfers.gcs_to_s3.GCSToS3Operator",
    },
    "google_api_to_s3_transfer": {
        "GoogleApiToS3Operator": (
            "airflow.providers.amazon.aws.transfers.google_api_to_s3.GoogleApiToS3Operator"
        ),
        "GoogleApiToS3Transfer": (
            "airflow.providers.amazon.aws.transfers.google_api_to_s3.GoogleApiToS3Operator"
        ),
    },
    "hive_operator": {
        "HiveOperator": "airflow.providers.apache.hive.operators.hive.HiveOperator",
    },
    "hive_stats_operator": {
        "HiveStatsCollectionOperator": (
            "airflow.providers.apache.hive.operators.hive_stats.HiveStatsCollectionOperator"
        ),
    },
    "hive_to_druid": {
        "HiveToDruidOperator": "airflow.providers.apache.druid.transfers.hive_to_druid.HiveToDruidOperator",
        "HiveToDruidTransfer": "airflow.providers.apache.druid.transfers.hive_to_druid.HiveToDruidOperator",
    },
    "hive_to_mysql": {
        "HiveToMySqlOperator": "airflow.providers.apache.hive.transfers.hive_to_mysql.HiveToMySqlOperator",
        "HiveToMySqlTransfer": "airflow.providers.apache.hive.transfers.hive_to_mysql.HiveToMySqlOperator",
    },
    "hive_to_samba_operator": {
        "HiveToSambaOperator": "airflow.providers.apache.hive.transfers.hive_to_samba.HiveToSambaOperator",
    },
    "http_operator": {
        "SimpleHttpOperator": "airflow.providers.http.operators.http.SimpleHttpOperator",
    },
    "jdbc_operator": {
        "JdbcOperator": "airflow.providers.jdbc.operators.jdbc.JdbcOperator",
    },
    "latest_only_operator": {
        "LatestOnlyOperator": "airflow.operators.latest_only.LatestOnlyOperator",
    },
    "mssql_operator": {
        "MsSqlOperator": "airflow.providers.microsoft.mssql.operators.mssql.MsSqlOperator",
    },
    "mssql_to_hive": {
        "MsSqlToHiveOperator": "airflow.providers.apache.hive.transfers.mssql_to_hive.MsSqlToHiveOperator",
        "MsSqlToHiveTransfer": "airflow.providers.apache.hive.transfers.mssql_to_hive.MsSqlToHiveOperator",
    },
    "mysql_operator": {
        "MySqlOperator": "airflow.providers.mysql.operators.mysql.MySqlOperator",
    },
    "mysql_to_hive": {
        "MySqlToHiveOperator": "airflow.providers.apache.hive.transfers.mysql_to_hive.MySqlToHiveOperator",
        "MySqlToHiveTransfer": "airflow.providers.apache.hive.transfers.mysql_to_hive.MySqlToHiveOperator",
    },
    "oracle_operator": {
        "OracleOperator": "airflow.providers.oracle.operators.oracle.OracleOperator",
    },
    "papermill_operator": {
        "PapermillOperator": "airflow.providers.papermill.operators.papermill.PapermillOperator",
    },
    "pig_operator": {
        "PigOperator": "airflow.providers.apache.pig.operators.pig.PigOperator",
    },
    "postgres_operator": {
        "Mapping": "airflow.providers.postgres.operators.postgres.Mapping",
        "PostgresOperator": "airflow.providers.postgres.operators.postgres.PostgresOperator",
    },
    "presto_check_operator": {
        "SQLCheckOperator": "airflow.providers.common.sql.operators.sql.SQLCheckOperator",
        "SQLIntervalCheckOperator": "airflow.providers.common.sql.operators.sql.SQLIntervalCheckOperator",
        "SQLValueCheckOperator": "airflow.providers.common.sql.operators.sql.SQLValueCheckOperator",
        "PrestoCheckOperator": "airflow.providers.common.sql.operators.sql.SQLCheckOperator",
        "PrestoIntervalCheckOperator": "airflow.providers.common.sql.operators.sql.SQLIntervalCheckOperator",
        "PrestoValueCheckOperator": "airflow.providers.common.sql.operators.sql.SQLValueCheckOperator",
    },
    "presto_to_mysql": {
        "PrestoToMySqlOperator": "airflow.providers.mysql.transfers.presto_to_mysql.PrestoToMySqlOperator",
        "PrestoToMySqlTransfer": "airflow.providers.mysql.transfers.presto_to_mysql.PrestoToMySqlOperator",
    },
    "python_operator": {
        "BranchPythonOperator": "airflow.operators.python.BranchPythonOperator",
        "PythonOperator": "airflow.operators.python.PythonOperator",
        "PythonVirtualenvOperator": "airflow.operators.python.PythonVirtualenvOperator",
        "ShortCircuitOperator": "airflow.operators.python.ShortCircuitOperator",
    },
    "redshift_to_s3_operator": {
        "RedshiftToS3Operator": "airflow.providers.amazon.aws.transfers.redshift_to_s3.RedshiftToS3Operator",
        "RedshiftToS3Transfer": "airflow.providers.amazon.aws.transfers.redshift_to_s3.RedshiftToS3Operator",
    },
    "s3_file_transform_operator": {
        "S3FileTransformOperator": (
            "airflow.providers.amazon.aws.operators.s3_file_transform.S3FileTransformOperator"
        ),
    },
    "s3_to_hive_operator": {
        "S3ToHiveOperator": "airflow.providers.apache.hive.transfers.s3_to_hive.S3ToHiveOperator",
        "S3ToHiveTransfer": "airflow.providers.apache.hive.transfers.s3_to_hive.S3ToHiveOperator",
    },
    "s3_to_redshift_operator": {
        "S3ToRedshiftOperator": "airflow.providers.amazon.aws.transfers.s3_to_redshift.S3ToRedshiftOperator",
        "S3ToRedshiftTransfer": "airflow.providers.amazon.aws.transfers.s3_to_redshift.S3ToRedshiftOperator",
    },
    "slack_operator": {
        "SlackAPIOperator": "airflow.providers.slack.operators.slack.SlackAPIOperator",
        "SlackAPIPostOperator": "airflow.providers.slack.operators.slack.SlackAPIPostOperator",
    },
    "sql": {
        "BaseSQLOperator": "airflow.providers.common.sql.operators.sql.BaseSQLOperator",
        "BranchSQLOperator": "airflow.providers.common.sql.operators.sql.BranchSQLOperator",
        "SQLCheckOperator": "airflow.providers.common.sql.operators.sql.SQLCheckOperator",
        "SQLColumnCheckOperator": "airflow.providers.common.sql.operators.sql.SQLColumnCheckOperator",
        "SQLIntervalCheckOperator": "airflow.providers.common.sql.operators.sql.SQLIntervalCheckOperator",
        "SQLTableCheckOperator": "airflow.providers.common.sql.operators.sql.SQLTableCheckOperator",
        "SQLThresholdCheckOperator": "airflow.providers.common.sql.operators.sql.SQLThresholdCheckOperator",
        "SQLValueCheckOperator": "airflow.providers.common.sql.operators.sql.SQLValueCheckOperator",
        "_convert_to_float_if_possible": (
            "airflow.providers.common.sql.operators.sql._convert_to_float_if_possible"
        ),
        "parse_boolean": "airflow.providers.common.sql.operators.sql.parse_boolean",
    },
    "sql_branch_operator": {
        "BranchSQLOperator": "airflow.providers.common.sql.operators.sql.BranchSQLOperator",
        "BranchSqlOperator": "airflow.providers.common.sql.operators.sql.BranchSQLOperator",
    },
    "sqlite_operator": {
        "SqliteOperator": "airflow.providers.sqlite.operators.sqlite.SqliteOperator",
    },
    "subdag_operator": {
        "SkippedStatePropagationOptions": "airflow.operators.subdag.SkippedStatePropagationOptions",
        "SubDagOperator": "airflow.operators.subdag.SubDagOperator",
    },
}

add_deprecated_classes(__deprecated_classes, __name__)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from stat import S_IFDIR, S_IFLNK, S_IFREG


class stat_result(dict):
    """
    stat_result: Result from stat, fstat, or lstat.

    This object provides a subset of os.stat_result attributes,
    for results returned from ObjectStoragePath.stat()

    It provides st_dev, st_ino, st_mode, st_nlink, st_uid, st_gid,
    st_size and st_mtime if they are available from the underlying
    storage. Extended attributes maybe accessed via dict access.

    See os.stat for more information.
    """

    st_dev = property(lambda self: 0)
    """device"""

    st_size = property(lambda self: self.get("size", 0))
    """total size, in bytes"""

    st_gid = property(lambda self: self.get("gid", 0))
    """group ID of owner"""

    st_uid = property(lambda self: self.get("uid", 0))
    """user ID of owner"""

    st_ino = property(lambda self: self.get("ino", 0))
    """inode"""

    st_nlink = property(lambda self: self.get("nlink", 0))
    """number of hard links"""

    @property
    def st_mtime(self):
        """Time of most recent content modification."""
        if "mtime" in self:
            return self.get("mtime")

        if "LastModified" in self:
            return self.get("LastModified").timestamp()

        # per posix.py
        return 0

    @property
    def st_mode(self):
        """Protection bits."""
        if "mode" in self:
            return self.get("mode")

        # per posix.py
        mode = 0o0
        if self.get("type", "") == "file":
            mode = S_IFREG

        if self.get("type", "") == "directory":
            mode = S_IFDIR

        if self.get("isLink", False):
            mode = S_IFLNK

        return mode



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
from functools import cached_property
from typing import TYPE_CHECKING, Callable

from flask import g
from flask_limiter import Limiter
from flask_limiter.util import get_remote_address
from sqlalchemy import select

from airflow.auth.managers.fab.security_manager.constants import EXISTING_ROLES as FAB_EXISTING_ROLES
from airflow.auth.managers.models.resource_details import (
    AccessView,
    ConnectionDetails,
    DagAccessEntity,
    DagDetails,
    PoolDetails,
    VariableDetails,
)
from airflow.auth.managers.utils.fab import (
    get_method_from_fab_action_map,
)
from airflow.exceptions import AirflowException
from airflow.models import Connection, DagRun, Pool, TaskInstance, Variable
from airflow.security.permissions import (
    ACTION_CAN_ACCESS_MENU,
    ACTION_CAN_READ,
    RESOURCE_ADMIN_MENU,
    RESOURCE_AUDIT_LOG,
    RESOURCE_BROWSE_MENU,
    RESOURCE_CLUSTER_ACTIVITY,
    RESOURCE_CONFIG,
    RESOURCE_CONNECTION,
    RESOURCE_DAG,
    RESOURCE_DAG_CODE,
    RESOURCE_DAG_DEPENDENCIES,
    RESOURCE_DAG_RUN,
    RESOURCE_DATASET,
    RESOURCE_DOCS,
    RESOURCE_DOCS_MENU,
    RESOURCE_JOB,
    RESOURCE_PLUGIN,
    RESOURCE_POOL,
    RESOURCE_PROVIDER,
    RESOURCE_SLA_MISS,
    RESOURCE_TASK_INSTANCE,
    RESOURCE_TASK_RESCHEDULE,
    RESOURCE_TRIGGER,
    RESOURCE_VARIABLE,
    RESOURCE_XCOM,
)
from airflow.utils.log.logging_mixin import LoggingMixin
from airflow.utils.session import NEW_SESSION, provide_session
from airflow.www.extensions.init_auth_manager import get_auth_manager
from airflow.www.utils import CustomSQLAInterface

EXISTING_ROLES = FAB_EXISTING_ROLES

if TYPE_CHECKING:
    from sqlalchemy.orm import Session

    from airflow.auth.managers.fab.models import Action, Resource
    from airflow.auth.managers.models.base_user import BaseUser


class AirflowSecurityManagerV2(LoggingMixin):
    """Custom security manager, which introduces a permission model adapted to Airflow.

    It's named V2 to differentiate it from the obsolete airflow.www.security.AirflowSecurityManager.
    """

    def __init__(self, appbuilder) -> None:
        super().__init__()
        self.appbuilder = appbuilder

        # Setup Flask-Limiter
        self.limiter = self.create_limiter()

        # Go and fix up the SQLAInterface used from the stock one to our subclass.
        # This is needed to support the "hack" where we had to edit
        # FieldConverter.conversion_table in place in airflow.www.utils
        for attr in dir(self):
            if attr.endswith("view"):
                view = getattr(self, attr, None)
                if view and getattr(view, "datamodel", None):
                    view.datamodel = CustomSQLAInterface(view.datamodel.obj)

    @staticmethod
    def before_request():
        """Run hook before request."""
        g.user = get_auth_manager().get_user()

    def create_limiter(self) -> Limiter:
        limiter = Limiter(key_func=get_remote_address)
        limiter.init_app(self.appbuilder.get_app)
        return limiter

    def register_views(self):
        """Allow auth managers to register their own views. By default, do nothing."""
        pass

    def has_access(
        self, action_name: str, resource_name: str, user=None, resource_pk: str | None = None
    ) -> bool:
        """
        Verify whether a given user could perform a certain action on the given resource.

        Example actions might include can_read, can_write, can_delete, etc.

        This function is called by FAB when accessing a view. See
        https://github.com/dpgaspar/Flask-AppBuilder/blob/c6fecdc551629e15467fde5d06b4437379d90592/flask_appbuilder/security/decorators.py#L134

        :param action_name: action_name on resource (e.g can_read, can_edit).
        :param resource_name: name of view-menu or resource.
        :param user: user
        :param resource_pk: the resource primary key (e.g. the connection ID)
        :return: Whether user could perform certain action on the resource.
        """
        if not user:
            user = g.user

        is_authorized_method = self._get_auth_manager_is_authorized_method(resource_name)
        return is_authorized_method(action_name, resource_pk, user)

    def create_admin_standalone(self) -> tuple[str | None, str | None]:
        """Perform the required steps when initializing airflow for standalone mode.

        If necessary, returns the username and password to be printed in the console for users to log in.
        """
        return None, None

    def add_limit_view(self, baseview):
        if not baseview.limits:
            return

        for limit in baseview.limits:
            self.limiter.limit(
                limit_value=limit.limit_value,
                key_func=limit.key_func,
                per_method=limit.per_method,
                methods=limit.methods,
                error_message=limit.error_message,
                exempt_when=limit.exempt_when,
                override_defaults=limit.override_defaults,
                deduct_when=limit.deduct_when,
                on_breach=limit.on_breach,
                cost=limit.cost,
            )(baseview.blueprint)

    @cached_property
    @provide_session
    def _auth_manager_is_authorized_map(
        self, session: Session = NEW_SESSION
    ) -> dict[str, Callable[[str, str | None, BaseUser | None], bool]]:
        """
        Return the map associating a FAB resource name to the corresponding auth manager is_authorized_ API.

        The function returned takes the FAB action name and the user as parameter.
        """
        auth_manager = get_auth_manager()
        methods = get_method_from_fab_action_map()

        def get_connection_id(resource_pk):
            if not resource_pk:
                return None
            connection = session.scalar(select(Connection).where(Connection.id == resource_pk).limit(1))
            if not connection:
                raise AirflowException("Connection not found")
            return connection.conn_id

        def get_dag_id_from_dagrun_id(resource_pk):
            if not resource_pk:
                return None
            dagrun = session.scalar(select(DagRun).where(DagRun.id == resource_pk).limit(1))
            if not dagrun:
                raise AirflowException("DagRun not found")
            return dagrun.dag_id

        def get_dag_id_from_task_instance(resource_pk):
            if not resource_pk:
                return None
            composite_pk = json.loads(resource_pk)
            ti = session.scalar(
                select(DagRun)
                .where(
                    TaskInstance.dag_id == composite_pk[0],
                    TaskInstance.task_id == composite_pk[1],
                    TaskInstance.run_id == composite_pk[2],
                    TaskInstance.map_index >= composite_pk[3],
                )
                .limit(1)
            )
            if not ti:
                raise AirflowException("Task instance not found")
            return ti.dag_id

        def get_pool_name(resource_pk):
            if not resource_pk:
                return None
            pool = session.scalar(select(Pool).where(Pool.id == resource_pk).limit(1))
            if not pool:
                raise AirflowException("Pool not found")
            return pool.pool

        def get_variable_key(resource_pk):
            if not resource_pk:
                return None
            variable = session.scalar(select(Variable).where(Variable.id == resource_pk).limit(1))
            if not variable:
                raise AirflowException("Connection not found")
            return variable.key

        return {
            RESOURCE_AUDIT_LOG: lambda action, resource_pk, user: auth_manager.is_authorized_dag(
                method=methods[action],
                access_entity=DagAccessEntity.AUDIT_LOG,
                user=user,
            ),
            RESOURCE_CLUSTER_ACTIVITY: lambda action, resource_pk, user: auth_manager.is_authorized_view(
                access_view=AccessView.CLUSTER_ACTIVITY,
                user=user,
            ),
            RESOURCE_CONFIG: lambda action, resource_pk, user: auth_manager.is_authorized_configuration(
                method=methods[action],
                user=user,
            ),
            RESOURCE_CONNECTION: lambda action, resource_pk, user: auth_manager.is_authorized_connection(
                method=methods[action],
                details=ConnectionDetails(conn_id=get_connection_id(resource_pk)),
                user=user,
            ),
            RESOURCE_DAG: lambda action, resource_pk, user: auth_manager.is_authorized_dag(
                method=methods[action],
                user=user,
            ),
            RESOURCE_DAG_CODE: lambda action, resource_pk, user: auth_manager.is_authorized_dag(
                method=methods[action],
                access_entity=DagAccessEntity.CODE,
                user=user,
            ),
            RESOURCE_DAG_DEPENDENCIES: lambda action, resource_pk, user: auth_manager.is_authorized_dag(
                method=methods[action],
                access_entity=DagAccessEntity.DEPENDENCIES,
                user=user,
            ),
            RESOURCE_DAG_RUN: lambda action, resource_pk, user: auth_manager.is_authorized_dag(
                method=methods[action],
                access_entity=DagAccessEntity.RUN,
                details=DagDetails(id=get_dag_id_from_dagrun_id(resource_pk)),
                user=user,
            ),
            RESOURCE_DATASET: lambda action, resource_pk, user: auth_manager.is_authorized_dataset(
                method=methods[action],
                user=user,
            ),
            RESOURCE_DOCS: lambda action, resource_pk, user: auth_manager.is_authorized_view(
                access_view=AccessView.DOCS,
                user=user,
            ),
            RESOURCE_PLUGIN: lambda action, resource_pk, user: auth_manager.is_authorized_view(
                access_view=AccessView.PLUGINS,
                user=user,
            ),
            RESOURCE_JOB: lambda action, resource_pk, user: auth_manager.is_authorized_view(
                access_view=AccessView.JOBS,
                user=user,
            ),
            RESOURCE_POOL: lambda action, resource_pk, user: auth_manager.is_authorized_pool(
                method=methods[action],
                details=PoolDetails(name=get_pool_name(resource_pk)),
                user=user,
            ),
            RESOURCE_PROVIDER: lambda action, resource_pk, user: auth_manager.is_authorized_view(
                access_view=AccessView.PROVIDERS,
                user=user,
            ),
            RESOURCE_SLA_MISS: lambda action, resource_pk, user: auth_manager.is_authorized_dag(
                method=methods[action],
                access_entity=DagAccessEntity.SLA_MISS,
                user=user,
            ),
            RESOURCE_TASK_INSTANCE: lambda action, resource_pk, user: auth_manager.is_authorized_dag(
                method=methods[action],
                access_entity=DagAccessEntity.TASK_INSTANCE,
                details=DagDetails(id=get_dag_id_from_task_instance(resource_pk)),
                user=user,
            ),
            RESOURCE_TASK_RESCHEDULE: lambda action, resource_pk, user: auth_manager.is_authorized_dag(
                method=methods[action],
                access_entity=DagAccessEntity.TASK_RESCHEDULE,
                user=user,
            ),
            RESOURCE_TRIGGER: lambda action, resource_pk, user: auth_manager.is_authorized_view(
                access_view=AccessView.TRIGGERS,
                user=user,
            ),
            RESOURCE_VARIABLE: lambda action, resource_pk, user: auth_manager.is_authorized_variable(
                method=methods[action],
                details=VariableDetails(key=get_variable_key(resource_pk)),
                user=user,
            ),
            RESOURCE_XCOM: lambda action, resource_pk, user: auth_manager.is_authorized_dag(
                method=methods[action],
                access_entity=DagAccessEntity.XCOM,
                user=user,
            ),
        }

    def _get_auth_manager_is_authorized_method(self, fab_resource_name: str) -> Callable:
        is_authorized_method = self._auth_manager_is_authorized_map.get(fab_resource_name)
        if is_authorized_method:
            return is_authorized_method
        elif fab_resource_name in [RESOURCE_DOCS_MENU, RESOURCE_ADMIN_MENU, RESOURCE_BROWSE_MENU]:
            # Display the "Browse", "Admin" and "Docs" dropdowns in the menu if the user has access to at
            # least one dropdown child
            return self._is_authorized_category_menu(fab_resource_name)
        else:
            # This means the page the user is trying to access is specific to the auth manager used
            # Example: the user list view in FabAuthManager
            return lambda action, resource_pk, user: get_auth_manager().is_authorized_custom_view(
                fab_action_name=ACTION_CAN_READ if action == ACTION_CAN_ACCESS_MENU else action,
                fab_resource_name=fab_resource_name,
                user=user,
            )

    def _is_authorized_category_menu(self, category: str) -> Callable:
        items = {item.name for item in self.appbuilder.menu.find(category).childs}
        return lambda action, resource_pk, user: any(
            self._get_auth_manager_is_authorized_method(fab_resource_name=item)(action, resource_pk, user)
            for item in items
        )

    """
    The following methods are specific to FAB auth manager. They still need to be "present" in the main
    security manager class, but they do nothing.
    """

    def get_action(self, name: str) -> Action:
        raise NotImplementedError("Only available if FAB auth manager is used")

    def get_resource(self, name: str) -> Resource:
        raise NotImplementedError("Only available if FAB auth manager is used")

    def add_permissions_view(self, base_action_names, resource_name):
        pass

    def add_permissions_menu(self, resource_name):
        pass



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Dag sub-commands."""
from __future__ import annotations

import ast
import errno
import json
import logging
import operator
import signal
import subprocess
import sys
import warnings
from typing import TYPE_CHECKING

from sqlalchemy import delete, select

from airflow import settings
from airflow.api.client import get_current_api_client
from airflow.api_connexion.schemas.dag_schema import dag_schema
from airflow.cli.simple_table import AirflowConsole
from airflow.configuration import conf
from airflow.exceptions import AirflowException, RemovedInAirflow3Warning
from airflow.jobs.job import Job
from airflow.models import DagBag, DagModel, DagRun, TaskInstance
from airflow.models.dag import DAG
from airflow.models.serialized_dag import SerializedDagModel
from airflow.utils import cli as cli_utils, timezone
from airflow.utils.cli import get_dag, get_dags, process_subdir, sigint_handler, suppress_logs_and_warning
from airflow.utils.dot_renderer import render_dag, render_dag_dependencies
from airflow.utils.providers_configuration_loader import providers_configuration_loaded
from airflow.utils.session import NEW_SESSION, create_session, provide_session
from airflow.utils.state import DagRunState

if TYPE_CHECKING:
    from graphviz.dot import Dot
    from sqlalchemy.orm import Session

    from airflow.timetables.base import DataInterval

log = logging.getLogger(__name__)


def _run_dag_backfill(dags: list[DAG], args) -> None:
    # If only one date is passed, using same as start and end
    args.end_date = args.end_date or args.start_date
    args.start_date = args.start_date or args.end_date

    run_conf = None
    if args.conf:
        run_conf = json.loads(args.conf)

    for dag in dags:
        if args.task_regex:
            dag = dag.partial_subset(
                task_ids_or_regex=args.task_regex, include_upstream=not args.ignore_dependencies
            )
            if not dag.task_dict:
                raise AirflowException(
                    f"There are no tasks that match '{args.task_regex}' regex. Nothing to run, exiting..."
                )

        if args.dry_run:
            print(f"Dry run of DAG {dag.dag_id} on {args.start_date}")
            dagrun_infos = dag.iter_dagrun_infos_between(earliest=args.start_date, latest=args.end_date)
            for dagrun_info in dagrun_infos:
                dr = DagRun(
                    dag.dag_id,
                    execution_date=dagrun_info.logical_date,
                    data_interval=dagrun_info.data_interval,
                )

                for task in dag.tasks:
                    print(f"Task {task.task_id} located in DAG {dag.dag_id}")
                    ti = TaskInstance(task, run_id=None)
                    ti.dag_run = dr
                    ti.dry_run()
        else:
            if args.reset_dagruns:
                DAG.clear_dags(
                    [dag],
                    start_date=args.start_date,
                    end_date=args.end_date,
                    confirm_prompt=not args.yes,
                    include_subdags=True,
                    dag_run_state=DagRunState.QUEUED,
                )

            try:
                dag.run(
                    start_date=args.start_date,
                    end_date=args.end_date,
                    mark_success=args.mark_success,
                    local=args.local,
                    donot_pickle=(args.donot_pickle or conf.getboolean("core", "donot_pickle")),
                    ignore_first_depends_on_past=args.ignore_first_depends_on_past,
                    ignore_task_deps=args.ignore_dependencies,
                    pool=args.pool,
                    delay_on_limit_secs=args.delay_on_limit,
                    verbose=args.verbose,
                    conf=run_conf,
                    rerun_failed_tasks=args.rerun_failed_tasks,
                    run_backwards=args.run_backwards,
                    continue_on_failures=args.continue_on_failures,
                    disable_retry=args.disable_retry,
                )
            except ValueError as vr:
                print(str(vr))
                sys.exit(1)


@cli_utils.action_cli
@providers_configuration_loaded
def dag_backfill(args, dag: list[DAG] | DAG | None = None) -> None:
    """Create backfill job or dry run for a DAG or list of DAGs using regex."""
    logging.basicConfig(level=settings.LOGGING_LEVEL, format=settings.SIMPLE_LOG_FORMAT)
    signal.signal(signal.SIGTERM, sigint_handler)
    warnings.warn(
        "--ignore-first-depends-on-past is deprecated as the value is always set to True",
        category=RemovedInAirflow3Warning,
    )

    if args.ignore_first_depends_on_past is False:
        args.ignore_first_depends_on_past = True

    if not args.start_date and not args.end_date:
        raise AirflowException("Provide a start_date and/or end_date")

    if not dag:
        dags = get_dags(args.subdir, dag_id=args.dag_id, use_regex=args.treat_dag_as_regex)
    elif isinstance(dag, list):
        dags = dag
    else:
        dags = [dag]
    del dag

    dags.sort(key=lambda d: d.dag_id)
    _run_dag_backfill(dags, args)
    if len(dags) > 1:
        log.info("All of the backfills are done.")


@cli_utils.action_cli
@providers_configuration_loaded
def dag_trigger(args) -> None:
    """Create a dag run for the specified dag."""
    api_client = get_current_api_client()
    try:
        message = api_client.trigger_dag(
            dag_id=args.dag_id,
            run_id=args.run_id,
            conf=args.conf,
            execution_date=args.exec_date,
            replace_microseconds=args.replace_microseconds,
        )
        AirflowConsole().print_as(
            data=[message] if message is not None else [],
            output=args.output,
        )
    except OSError as err:
        raise AirflowException(err)


@cli_utils.action_cli
@providers_configuration_loaded
def dag_delete(args) -> None:
    """Delete all DB records related to the specified dag."""
    api_client = get_current_api_client()
    if (
        args.yes
        or input("This will drop all existing records related to the specified DAG. Proceed? (y/n)").upper()
        == "Y"
    ):
        try:
            message = api_client.delete_dag(dag_id=args.dag_id)
            print(message)
        except OSError as err:
            raise AirflowException(err)
    else:
        print("Cancelled")


@cli_utils.action_cli
@providers_configuration_loaded
def dag_pause(args) -> None:
    """Pauses a DAG."""
    set_is_paused(True, args)


@cli_utils.action_cli
@providers_configuration_loaded
def dag_unpause(args) -> None:
    """Unpauses a DAG."""
    set_is_paused(False, args)


@providers_configuration_loaded
def set_is_paused(is_paused: bool, args) -> None:
    """Set is_paused for DAG by a given dag_id."""
    dag = DagModel.get_dagmodel(args.dag_id)

    if not dag:
        raise SystemExit(f"DAG: {args.dag_id} does not exist in 'dag' table")

    dag.set_is_paused(is_paused=is_paused)

    print(f"Dag: {args.dag_id}, paused: {is_paused}")


@providers_configuration_loaded
def dag_dependencies_show(args) -> None:
    """Display DAG dependencies, save to file or show as imgcat image."""
    dot = render_dag_dependencies(SerializedDagModel.get_dag_dependencies())
    filename = args.save
    imgcat = args.imgcat

    if filename and imgcat:
        raise SystemExit(
            "Option --save and --imgcat are mutually exclusive. "
            "Please remove one option to execute the command.",
        )
    elif filename:
        _save_dot_to_file(dot, filename)
    elif imgcat:
        _display_dot_via_imgcat(dot)
    else:
        print(dot.source)


@providers_configuration_loaded
def dag_show(args) -> None:
    """Display DAG or saves its graphic representation to the file."""
    dag = get_dag(args.subdir, args.dag_id)
    dot = render_dag(dag)
    filename = args.save
    imgcat = args.imgcat

    if filename and imgcat:
        raise SystemExit(
            "Option --save and --imgcat are mutually exclusive. "
            "Please remove one option to execute the command.",
        )
    elif filename:
        _save_dot_to_file(dot, filename)
    elif imgcat:
        _display_dot_via_imgcat(dot)
    else:
        print(dot.source)


def _display_dot_via_imgcat(dot: Dot) -> None:
    data = dot.pipe(format="png")
    try:
        with subprocess.Popen("imgcat", stdout=subprocess.PIPE, stdin=subprocess.PIPE) as proc:
            out, err = proc.communicate(data)
            if out:
                print(out.decode("utf-8"))
            if err:
                print(err.decode("utf-8"))
    except OSError as e:
        if e.errno == errno.ENOENT:
            raise SystemExit("Failed to execute. Make sure the imgcat executables are on your systems 'PATH'")
        else:
            raise


def _save_dot_to_file(dot: Dot, filename: str) -> None:
    filename_without_ext, _, ext = filename.rpartition(".")
    dot.render(filename=filename_without_ext, format=ext, cleanup=True)
    print(f"File {filename} saved")


@cli_utils.action_cli
@providers_configuration_loaded
@provide_session
def dag_state(args, session: Session = NEW_SESSION) -> None:
    """
    Return the state (and conf if exists) of a DagRun at the command line.

    >>> airflow dags state tutorial 2015-01-01T00:00:00.000000
    running
    >>> airflow dags state a_dag_with_conf_passed 2015-01-01T00:00:00.000000
    failed, {"name": "bob", "age": "42"}
    """
    dag = DagModel.get_dagmodel(args.dag_id, session=session)

    if not dag:
        raise SystemExit(f"DAG: {args.dag_id} does not exist in 'dag' table")
    dr = session.scalar(select(DagRun).filter_by(dag_id=args.dag_id, execution_date=args.execution_date))
    out = dr.state if dr else None
    conf_out = ""
    if out and dr.conf:
        conf_out = ", " + json.dumps(dr.conf)
    print(str(out) + conf_out)


@cli_utils.action_cli
@providers_configuration_loaded
def dag_next_execution(args) -> None:
    """
    Return the next execution datetime of a DAG at the command line.

    >>> airflow dags next-execution tutorial
    2018-08-31 10:38:00
    """
    dag = get_dag(args.subdir, args.dag_id)

    with create_session() as session:
        last_parsed_dag: DagModel = session.scalars(
            select(DagModel).where(DagModel.dag_id == dag.dag_id)
        ).one()

    if last_parsed_dag.get_is_paused():
        print("[INFO] Please be reminded this DAG is PAUSED now.", file=sys.stderr)

    def print_execution_interval(interval: DataInterval | None):
        if interval is None:
            print(
                "[WARN] No following schedule can be found. "
                "This DAG may have schedule interval '@once' or `None`.",
                file=sys.stderr,
            )
            print(None)
            return
        print(interval.start.isoformat())

    next_interval = dag.get_next_data_interval(last_parsed_dag)
    print_execution_interval(next_interval)

    for _ in range(1, args.num_executions):
        next_info = dag.next_dagrun_info(next_interval, restricted=False)
        next_interval = None if next_info is None else next_info.data_interval
        print_execution_interval(next_interval)


@cli_utils.action_cli
@suppress_logs_and_warning
@providers_configuration_loaded
def dag_list_dags(args) -> None:
    """Display dags with or without stats at the command line."""
    dagbag = DagBag(process_subdir(args.subdir))
    if dagbag.import_errors:
        from rich import print as rich_print

        rich_print(
            "[red][bold]Error:[/bold] Failed to load all files. "
            "For details, run `airflow dags list-import-errors`",
            file=sys.stderr,
        )
    AirflowConsole().print_as(
        data=sorted(dagbag.dags.values(), key=operator.attrgetter("dag_id")),
        output=args.output,
        mapper=lambda x: {
            "dag_id": x.dag_id,
            "filepath": x.filepath,
            "owner": x.owner,
            "paused": x.get_is_paused(),
        },
    )


@cli_utils.action_cli
@suppress_logs_and_warning
@providers_configuration_loaded
@provide_session
def dag_details(args, session=NEW_SESSION):
    """Get DAG details given a DAG id."""
    dag = DagModel.get_dagmodel(args.dag_id, session=session)
    if not dag:
        raise SystemExit(f"DAG: {args.dag_id} does not exist in 'dag' table")
    dag_detail = dag_schema.dump(dag)

    if args.output in ["table", "plain"]:
        data = [{"property_name": key, "property_value": value} for key, value in dag_detail.items()]
    else:
        data = [dag_detail]

    AirflowConsole().print_as(
        data=data,
        output=args.output,
    )


@cli_utils.action_cli
@suppress_logs_and_warning
@providers_configuration_loaded
def dag_list_import_errors(args) -> None:
    """Display dags with import errors on the command line."""
    dagbag = DagBag(process_subdir(args.subdir))
    data = []
    for filename, errors in dagbag.import_errors.items():
        data.append({"filepath": filename, "error": errors})
    AirflowConsole().print_as(
        data=data,
        output=args.output,
    )
    if data:
        sys.exit(1)


@cli_utils.action_cli
@suppress_logs_and_warning
@providers_configuration_loaded
def dag_report(args) -> None:
    """Display dagbag stats at the command line."""
    dagbag = DagBag(process_subdir(args.subdir))
    AirflowConsole().print_as(
        data=dagbag.dagbag_stats,
        output=args.output,
        mapper=lambda x: {
            "file": x.file,
            "duration": x.duration,
            "dag_num": x.dag_num,
            "task_num": x.task_num,
            "dags": sorted(ast.literal_eval(x.dags)),
        },
    )


@cli_utils.action_cli
@suppress_logs_and_warning
@providers_configuration_loaded
@provide_session
def dag_list_jobs(args, dag: DAG | None = None, session: Session = NEW_SESSION) -> None:
    """List latest n jobs."""
    queries = []
    if dag:
        args.dag_id = dag.dag_id
    if args.dag_id:
        dag = DagModel.get_dagmodel(args.dag_id, session=session)

        if not dag:
            raise SystemExit(f"DAG: {args.dag_id} does not exist in 'dag' table")
        queries.append(Job.dag_id == args.dag_id)

    if args.state:
        queries.append(Job.state == args.state)

    fields = ["dag_id", "state", "job_type", "start_date", "end_date"]
    all_jobs_iter = session.scalars(
        select(Job).where(*queries).order_by(Job.start_date.desc()).limit(args.limit)
    )
    all_jobs = [{f: str(job.__getattribute__(f)) for f in fields} for job in all_jobs_iter]

    AirflowConsole().print_as(
        data=all_jobs,
        output=args.output,
    )


@cli_utils.action_cli
@suppress_logs_and_warning
@providers_configuration_loaded
@provide_session
def dag_list_dag_runs(args, dag: DAG | None = None, session: Session = NEW_SESSION) -> None:
    """List dag runs for a given DAG."""
    if dag:
        args.dag_id = dag.dag_id
    else:
        dag = DagModel.get_dagmodel(args.dag_id, session=session)

        if not dag:
            raise SystemExit(f"DAG: {args.dag_id} does not exist in 'dag' table")

    state = args.state.lower() if args.state else None
    dag_runs = DagRun.find(
        dag_id=args.dag_id,
        state=state,
        no_backfills=args.no_backfill,
        execution_start_date=args.start_date,
        execution_end_date=args.end_date,
        session=session,
    )

    dag_runs.sort(key=lambda x: x.execution_date, reverse=True)
    AirflowConsole().print_as(
        data=dag_runs,
        output=args.output,
        mapper=lambda dr: {
            "dag_id": dr.dag_id,
            "run_id": dr.run_id,
            "state": dr.state,
            "execution_date": dr.execution_date.isoformat(),
            "start_date": dr.start_date.isoformat() if dr.start_date else "",
            "end_date": dr.end_date.isoformat() if dr.end_date else "",
        },
    )


@cli_utils.action_cli
@providers_configuration_loaded
@provide_session
def dag_test(args, dag: DAG | None = None, session: Session = NEW_SESSION) -> None:
    """Execute one single DagRun for a given DAG and execution date."""
    run_conf = None
    if args.conf:
        try:
            run_conf = json.loads(args.conf)
        except ValueError as e:
            raise SystemExit(f"Configuration {args.conf!r} is not valid JSON. Error: {e}")
    execution_date = args.execution_date or timezone.utcnow()
    dag = dag or get_dag(subdir=args.subdir, dag_id=args.dag_id)
    dag.test(execution_date=execution_date, run_conf=run_conf, session=session)
    show_dagrun = args.show_dagrun
    imgcat = args.imgcat_dagrun
    filename = args.save_dagrun
    if show_dagrun or imgcat or filename:
        tis = session.scalars(
            select(TaskInstance).where(
                TaskInstance.dag_id == args.dag_id,
                TaskInstance.execution_date == execution_date,
            )
        ).all()

        dot_graph = render_dag(dag, tis=tis)
        print()
        if filename:
            _save_dot_to_file(dot_graph, filename)
        if imgcat:
            _display_dot_via_imgcat(dot_graph)
        if show_dagrun:
            print(dot_graph.source)


@cli_utils.action_cli
@providers_configuration_loaded
@provide_session
def dag_reserialize(args, session: Session = NEW_SESSION) -> None:
    """Serialize a DAG instance."""
    session.execute(delete(SerializedDagModel).execution_options(synchronize_session=False))

    if not args.clear_only:
        dagbag = DagBag(process_subdir(args.subdir))
        dagbag.sync_to_db(session=session)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

from airflow.io import get_fs, has_fs
from airflow.utils.module_loading import qualname

if TYPE_CHECKING:
    from fsspec import AbstractFileSystem

    from airflow.io.typedef import Properties


class ObjectStore:
    """Manages a filesystem or object storage."""

    __version__: ClassVar[int] = 1

    method: str
    conn_id: str | None
    protocol: str
    storage_options: Properties | None

    _fs: AbstractFileSystem | None = None

    def __init__(
        self,
        protocol: str,
        conn_id: str | None,
        fs: AbstractFileSystem | None = None,
        storage_options: Properties | None = None,
    ):
        self.conn_id = conn_id
        self.protocol = protocol
        self._fs = fs
        self.storage_options = storage_options

    def __str__(self):
        return f"{self.protocol}-{self.conn_id}" if self.conn_id else self.protocol

    @property
    def fs(self) -> AbstractFileSystem:
        return self._connect()

    @property
    def fsid(self) -> str:
        """
        Get the filesystem id for this store in order to be able to compare across instances.

        The underlying `fsid` is returned from the filesystem if available, otherwise it is generated
        from the protocol and connection ID.

        :return: deterministic the filesystem ID
        """
        fs = self._connect()
        try:
            return fs.fsid
        except NotImplementedError:
            return f"{self.fs.protocol}-{self.conn_id or 'env'}"

    def serialize(self):
        return {
            "protocol": self.protocol,
            "conn_id": self.conn_id,
            "filesystem": qualname(self._fs) if self._fs else None,
            "storage_options": self.storage_options,
        }

    @classmethod
    def deserialize(cls, data: dict[str, str], version: int):
        if version > cls.__version__:
            raise ValueError(f"Cannot deserialize version {version} for {cls.__name__}")

        protocol = data["protocol"]
        conn_id = data["conn_id"]

        alias = f"{protocol}-{conn_id}" if conn_id else protocol

        if store := _STORE_CACHE.get(alias):
            return store

        if not has_fs(protocol) and "filesystem" in data and data["filesystem"]:
            raise ValueError(
                f"No attached filesystem found for {data['filesystem']} with "
                f"protocol {data['protocol']}. Please use attach() for this protocol and filesystem."
            )

        return attach(protocol=protocol, conn_id=conn_id, storage_options=data["storage_options"])

    def _connect(self) -> AbstractFileSystem:
        if self._fs is None:
            self._fs = get_fs(self.protocol, self.conn_id)
        return self._fs

    def __eq__(self, other):
        return isinstance(other, type(self)) and other.conn_id == self.conn_id and other._fs == self._fs


_STORE_CACHE: dict[str, ObjectStore] = {}


def attach(
    protocol: str | None = None,
    conn_id: str | None = None,
    alias: str | None = None,
    encryption_type: str | None = "",
    fs: AbstractFileSystem | None = None,
    **kwargs,
) -> ObjectStore:
    """
    Attach a filesystem or object storage.

    :param alias: the alias to be used to refer to the store, autogenerated if omitted
    :param protocol: the scheme that is used without ://
    :param conn_id: the connection to use to connect to the filesystem
    :param encryption_type: the encryption type to use to connect to the filesystem
    :param fs: the filesystem type to use to connect to the filesystem
    """
    if alias:
        if store := _STORE_CACHE.get(alias):
            return store
        elif not protocol:
            raise ValueError(f"No registered store with alias: {alias}")

    if not protocol:
        raise ValueError("No protocol specified and no alias provided")

    if not alias:
        alias = f"{protocol}-{conn_id}" if conn_id else protocol
        if store := _STORE_CACHE.get(alias):
            return store

    _STORE_CACHE[alias] = store = ObjectStore(protocol=protocol, conn_id=conn_id, fs=fs, **kwargs)

    return store



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Scheduler command."""
from __future__ import annotations

import logging
from argparse import Namespace
from contextlib import contextmanager
from multiprocessing import Process

from airflow import settings
from airflow.api_internal.internal_api_call import InternalApiConfig
from airflow.cli.commands.daemon_utils import run_command_with_daemon_option
from airflow.configuration import conf
from airflow.executors.executor_loader import ExecutorLoader
from airflow.jobs.job import Job, run_job
from airflow.jobs.scheduler_job_runner import SchedulerJobRunner
from airflow.utils import cli as cli_utils
from airflow.utils.cli import process_subdir
from airflow.utils.providers_configuration_loader import providers_configuration_loaded
from airflow.utils.scheduler_health import serve_health_check

log = logging.getLogger(__name__)


def _run_scheduler_job(args) -> None:
    job_runner = SchedulerJobRunner(
        job=Job(), subdir=process_subdir(args.subdir), num_runs=args.num_runs, do_pickle=args.do_pickle
    )
    ExecutorLoader.validate_database_executor_compatibility(job_runner.job.executor)
    InternalApiConfig.force_database_direct_access()
    enable_health_check = conf.getboolean("scheduler", "ENABLE_HEALTH_CHECK")
    with _serve_logs(args.skip_serve_logs), _serve_health_check(enable_health_check):
        try:
            run_job(job=job_runner.job, execute_callable=job_runner._execute)
        except Exception:
            log.exception("Exception when running scheduler job")


@cli_utils.action_cli
@providers_configuration_loaded
def scheduler(args: Namespace):
    """Start Airflow Scheduler."""
    print(settings.HEADER)

    run_command_with_daemon_option(
        args=args,
        process_name="scheduler",
        callback=lambda: _run_scheduler_job(args),
        should_setup_logging=True,
    )


@contextmanager
def _serve_logs(skip_serve_logs: bool = False):
    """Start serve_logs sub-process."""
    from airflow.utils.serve_logs import serve_logs

    sub_proc = None
    executor_class, _ = ExecutorLoader.import_default_executor_cls()
    if executor_class.serve_logs:
        if skip_serve_logs is False:
            sub_proc = Process(target=serve_logs)
            sub_proc.start()
    yield
    if sub_proc:
        sub_proc.terminate()


@contextmanager
def _serve_health_check(enable_health_check: bool = False):
    """Start serve_health_check sub-process."""
    sub_proc = None
    if enable_health_check:
        sub_proc = Process(target=serve_health_check)
        sub_proc.start()
    yield
    if sub_proc:
        sub_proc.terminate()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import contextlib
import logging
import traceback
from typing import TYPE_CHECKING

from airflow.utils.session import create_session

if TYPE_CHECKING:
    from sqlalchemy.orm import Session

__current_task_instance_session: Session | None = None

log = logging.getLogger(__name__)


def get_current_task_instance_session() -> Session:
    global __current_task_instance_session
    if not __current_task_instance_session:
        log.warning("No task session set for this task. Continuing but this likely causes a resource leak.")
        log.warning("Please report this and stacktrace below to https://github.com/apache/airflow/issues")
        for filename, line_number, name, line in traceback.extract_stack():
            log.warning('File: "%s", %s , in %s', filename, line_number, name)
            if line:
                log.warning("  %s", line.strip())
        __current_task_instance_session = create_session()
    return __current_task_instance_session


@contextlib.contextmanager
def set_current_task_instance_session(session: Session):
    global __current_task_instance_session
    if __current_task_instance_session:
        raise RuntimeError(
            "Session already set for this task. "
            "You can only have one 'set_current_task_session' context manager active at a time."
        )
    __current_task_instance_session = session
    try:
        yield
    finally:
        __current_task_instance_session = None



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import signal
from argparse import Namespace
from typing import Callable

from daemon import daemon
from daemon.pidfile import TimeoutPIDLockFile

from airflow import settings
from airflow.utils.cli import setup_locations, setup_logging, sigint_handler, sigquit_handler
from airflow.utils.process_utils import check_if_pidfile_process_is_running


def run_command_with_daemon_option(
    *,
    args: Namespace,
    process_name: str,
    callback: Callable,
    should_setup_logging: bool = False,
    umask: str = settings.DAEMON_UMASK,
    pid_file: str | None = None,
):
    """Run the command in a daemon process if daemon mode enabled or within this process if not.

    :param args: the set of arguments passed to the original CLI command
    :param process_name: process name used in naming log and PID files for the daemon
    :param callback: the actual command to run with or without daemon context
    :param should_setup_logging: if true, then a log file handler for the daemon process will be created
    :param umask: file access creation mask ("umask") to set for the process on daemon start
    :param pid_file: if specified, this file path us used to store daemon process PID.
        If not specified, a file path is generated with the default pattern.
    """
    if args.daemon:
        pid, stdout, stderr, log_file = setup_locations(
            process=process_name, stdout=args.stdout, stderr=args.stderr, log=args.log_file
        )
        if pid_file:
            pid = pid_file

        # Check if the process is already running; if not but a pidfile exists, clean it up
        check_if_pidfile_process_is_running(pid_file=pid, process_name=process_name)

        if should_setup_logging:
            files_preserve = [setup_logging(log_file)]
        else:
            files_preserve = None
        with open(stdout, "a") as stdout_handle, open(stderr, "a") as stderr_handle:
            stdout_handle.truncate(0)
            stderr_handle.truncate(0)

            ctx = daemon.DaemonContext(
                pidfile=TimeoutPIDLockFile(pid, -1),
                files_preserve=files_preserve,
                stdout=stdout_handle,
                stderr=stderr_handle,
                umask=int(umask, 8),
            )

            with ctx:
                # in daemon context stats client needs to be reinitialized.
                from airflow.stats import Stats

                Stats.instance = None
                callback()
    else:
        signal.signal(signal.SIGINT, sigint_handler)
        signal.signal(signal.SIGTERM, sigint_handler)
        signal.signal(signal.SIGQUIT, sigquit_handler)
        callback()
